/*****************************************************************************
*                                                                            *
*     VITO -- Verilog Implicit To One-hot translator                         *
*     $Id: legal.txt,v 1.7 1999/06/03 15:24:08 mwinkel Apr_19_2004_Version_1.5 $
*                                                                            *
*     This software translates implicit style Verilog into a form suitable   *
*     for most commercial synthesis tools.  The translation is achieved      *
*     by constructing a one-hot machine for the ASM described in the         *
*     implicit style Verilog.                                                *
*                                                                            *
*     This work is being performed at the University of Wyoming and SUNY     *
*     College at Brockport.  For more information:                           *
*	   www.verilog.vito.com                                              *
*          http://strawberry.uwyo.edu/~vito                                  *
*          http://cs.brockport.edu/~jshuler/vito                             *
*                                                                            *
*     Authors:                                                               *
*          James D. Shuler  jshuler@cs.brockport.edu                         *
*          Mark G. Arnold   marnold@uwyo.edu                                 *
*                                                                            *
*                         NOTICE AND DISCLAIMER                              *
*                                                                            *
* No use of this software is authorized except under this NOTICE AND         *
* DISCLAIMER.  Downloading, use, copying, modification, and/or               *
* distribution of this software in its original or modified form implies     *
* your acceptance of this NOTICE and DISCLAIMER.                             *
*                                                                            *
* Permission to use, copy, modify and distribute this software and its       *
* documentation, for any purpose and without fee is hereby granted,          *
* provided that this permission notice appears prominently in supporting     *
* documentation and must be viewed prior to use, copying, modification or    *
* distribution.  You are responsible for any modifications to the software   *
* which you make and notice that the software has been modified must be      *
* appended to this notice prior to further use, copying, modification or     *
* distribution.                                                              *
*                                                                            *
* This software is provided AS IS without warranty of any kind,              *
* including without limitation the warranties that the software is           *
* non-infringing, merchantable, or fit for a particular purpose, including   *
* high risk activities.  The entire risk as to the quality and performance   *
* of the software is born by you.  Should the software prove defective in    *
* any respect, you and not the developers, nor any parties associated with   *
* the developers, assume the entire cost of  any service and repair.         *
*                                                                            *
* This software may be subject to the Export Control Laws of the United      *
* States of America.  It is your responsibility to determine the             *
* applicable laws and regulations and comply with them.                      *
*                                                                            *
*****************************************************************************/
/*****************************************************************************
$Id: vito.c,v 1.7 1999/06/03 17:07:15 mwinkel Apr_19_2004_Version_1.5 $

	Main driving code and utility functions of VITO.

*****************************************************************************/
#include "vito.h"
#include <stdarg.h>
#include <stdlib.h>

#define out_filename		rc_string[CFFILE]

static const char rcsid[] = "$Id: vito.c,v 1.7 1999/06/03 17:07:15 mwinkel Apr_19_2004_Version_1.5 $"; /* Rev number included in executable */

extern char *yytext;	/* last token read in by Lex */
int ignore_comments;	/* TRUE if comments should be thrown away.  Used during
			   processing of "valid" always blocks */
int ai_process;		/* TRUE if processing a "valid" always block */
int initial_begin,always_begin;	/* TRUE if this is the start of an initial or always block */
put_type token_print,	/* The place to put tokens when tp is called */
         tp_other;	/* The second place to put tokens when tp is called */
int do_new_wire;	/* TRUE if replicating a reg declaration to make wires */
int duplicate_regs;	/* TRUE if wire declarations are to be made in .arch */
int last_source_file;	/* 1+end of source file list */
char *source_files[MAX_NO_FILES];	/* list of source files */
char *rcfile_name;	/* name of the rcfile */
char *outfile_name;	/* name of the output file */
char *rc_string[MAXRC]={	/* Array of initialization strings */
			  "vito.out",
			  "vito.stmt",
			  "vito.arch",
			  "",
			  "s_",
			  "sT_",
			  "tmp_",
			  "join_",
			  "ff_",
			  "qual_",
			  "reset",
			  "sysclk",
			  "@(posedge sysclk)",
			  "new_",
			  "@(posedge sysclk or posedge reset)",
			  "reset",
			  "vito.tail",
			  ""
			};
char token_string[MAX_TOKEN_STRING];	/* a string of tokens read in so far since 
					  it was last cleared */
int ts_length;		/* the length of token_string */
int line_no;		/* The current line number of the input file */
int no_modules;		/* number of modules encountered */
int current_module;	/* number of current module */
int implicit_module;	/* TRUE if this module contains implicit style block */
module_info_type module_info[MAX_NO_MODULES];
int no_errors,		/* The current number of errors issued */
    no_warnings;	/* the current number of warnings issued */
int reset_ID,clock_ID;	/* index to ID_table for reset and clock names */
int reset_in_port,clock_in_port;	/* TRUE if reset or clock appears in module port list */
int last_ID;		/* the next free entry in the ID_table */
int initial_last_ID;	/* the starting value for last_ID at the start of each module */
char *ID_table[MAXID];	/* table of identifiers in the source code */
char *library_paths;	/* list of paths to library directories */
int verbose_mode;	/* TRUE if verbose mode is on */
int debug_mode;		/* TRUE if debugging mode is on */
int table_mode;		/* TRUE if printing table mode is on */
int preprocess_mode;	/* TRUE if only the prepocessor step is to be performed */
case_stack_type *case_stack;	/* stack for storing case expressions--used in converting into
				   nested IFs */
dynamic_istack_type *token_print_stack;	/* stack for storing token_print values */
comment_type *current_comment,	/* first node in linked list of comments */
	     *current_comment_end;	/* last node in linked list of comments */
FILE *stmt_file;	/* the file used with PUT_STMT */
FILE *out_file;		/* the file used with PUT_OUT */
FILE *arch_file;	/* the file used with PUT_ARCH */
FILE *tail_file;	/* the file used with PUT_TAIL */

/*****************************************************************************

			y y w r a p

	 Lex/YACC function to obtain next file for input.

*****************************************************************************/
int yywrap(void)
{
   return TRUE;
}
/*****************************************************************************

			c o n c a t

      Concatenates the given string onto the end of token_string.
   ts_length is updated with the length of the new string.

Input:	string = the string to append to token_string

*****************************************************************************/
void concat(char *string)
{
   int length;

   length=strlen(string);
   if (ts_length+length >= MAX_TOKEN_STRING) {
      length = MAX_TOKEN_STRING - ts_length - 1;
      error("Exceeded internal string.  Increase MAX_TOKEN_STRING.\n");
   }
   ts_length += length;
   strncat(token_string,string,length);
}
/*****************************************************************************

			t s _ c l e a r

      Clears token_string.  Typically performed at the beginnings of lines and
other important spots in the input code.

*****************************************************************************/
void ts_clear(void)
{
   token_string[0] = '\0';
   ts_length=0;
}
/*****************************************************************************

			G e t _ I D

	 Returns the identifier string located in the ID_table at index
   location.

Input:	location = index of identifier in table
Output:	Get_ID = identifier string.

*****************************************************************************/
char *Get_ID(int location)
{
   if (location>=0 && location<last_ID)
      return ID_table[location];
   error("Internal error.  Bad location in Get_ID.\n");
}
/*****************************************************************************

			F i n d _ I D

	 Finds the location of an identifier in the ID_table given the
   the identifiers string.  If it is not found, this returns -1.

Input:	ID = the identifier to look up
Output:	Find_ID = the index in the ID_table

*****************************************************************************/
int Find_ID(char *ID)
{
   int cur_loc;
   for(cur_loc=0;cur_loc<last_ID;cur_loc++)
      if(strcmp(ID,ID_table[cur_loc])==0)
	 return cur_loc;
   return -1;
}
/*****************************************************************************

			I n s e r t _ I D

	 Inserts the given identifier into the ID_table and returns the
   location of where it is inserted.

Input:	ID = string for identifier
	length = the length of the identifier string (without \0)
Output:	Insert_ID = the location of the inserted identifier

*****************************************************************************/
int Insert_ID(char *ID,int length)
{
   int loc;
   if (last_ID >= MAXID-1) {
      error("Exceeded internal ID table.  Increase MAXID.`n");
      exit(-1);
   }
   /* see if already in table */
   loc=Find_ID(ID);
   if (loc>=0)
      return loc;
   /* if not, put ID in the table */
   ID_table[last_ID] = (char *)malloc(length+1);
   if (ID_table[last_ID]==NULL) {
      error("Malloc returned NULL.  Ran out of memory.\n");
      exit(EXIT_FAILURE);
   }
   strcpy(ID_table[last_ID],ID);
   last_ID++;
   return (last_ID-1);
}
/************************************************************************

			r e m o v e _ I D

	Removes all the identifiers in the ID table starting with the
   given position.  last_ID is updated to the start of the cleaned area.

input:	start = the starting position in table to remove from

************************************************************************/
void remove_ID(int start)
{
   int cur_ID;

   for (cur_ID=start;cur_ID<last_ID;cur_ID++)
      free(ID_table[cur_ID]);
   last_ID = start;
}
/*****************************************************************************

			i n i t _ r c _ s t r i n g

	Initializes one string in the rc_string table to the given value.

Input:	line = the index of the string
	string = the value of the string

*****************************************************************************/
void init_rc_string(int line,char *string)
{
   rc_string[line] = (char *)malloc(strlen(string)+1);
   if (rc_string[line]==NULL) {
      error("Malloc returned NULL.  Ran out of memory.\n");
      exit(EXIT_FAILURE);
   }
   strcpy(rc_string[line],string);
}
/*****************************************************************************

			r e a d _ r c _ f i l e

	 Initializes the rc_string table.  This contains the user configurable
values for our tool.  If the file RC_FILE exists, it is used as the values.
If it does not exist, then a predefined set of values is used.

*****************************************************************************/
void read_rc_file(void)
{
   FILE *rc_file;
   int cur_line,length;
   char string[MAXSTRING];

   cur_line=0;
   rc_file = fopen(rcfile_name,"r");
   if (rc_file!=NULL) {
      /* read strings from the file */
      while (!feof(rc_file) && cur_line<MAXRC) {
         fgets(string,MAXSTRING,rc_file);
	 if (feof(rc_file))
	    break;
	 length = strlen(string);
	 if (length<=0)
	    continue;
	 if (string[0]==RC_COMMENT)
	    continue;
	 if (string[length-1]=='\n')
	    string[length-1] = '\0';
	 else
	    error("Line too long in '%s' file.  Increase MAXSTRING.",rcfile_name);
	 init_rc_string(cur_line,string);
         cur_line++;
      }
      fclose(rc_file);
   }
   else {
      /* rcfile_name does not exist */
      if (verbose_mode)
	 info("Unable to open '%s'.  Using defaults.\n",rcfile_name);
   }
   /* Put the reset and clock names into ID_table for future matching */
   reset_ID = Insert_ID(rc_string[RESET],strlen(rc_string[RESET]));
   clock_ID = Insert_ID(rc_string[CLOCK],strlen(rc_string[CLOCK]));
}
/*****************************************************************************

			e r r o r

	 Prints a uniform error message giving the name of the tool and the
   line number of the error.  If there are too many errors, then this
   program is exited.

Input:	string = the error message
	... = arguments for the error string

*****************************************************************************/
void error(char *string, ...)
{
   va_list ap;
   va_start(ap,string);
   fprintf(stderr,"%s Error %d:  ",NAME_OF_TOOL,line_no);
   vfprintf(stderr,string,ap);
   no_errors++;
   if (no_errors > MAXNOERRORS) {
      fprintf(stderr,"%s:  Too many errors.  Goodbye.\n",NAME_OF_TOOL);
      exit(EXIT_FAILURE);
   }
   ts_clear();
}
/*****************************************************************************

			w a r n i n g

	 Prints a uniform warning message giving the name of the tool and the
   line number of the error.

Input:	string = the warning message
	... = arguments for the warning string

*****************************************************************************/
void warning(char *string, ...)
{
   va_list ap;
   va_start(ap,string);
   fprintf(stderr,"%s Warning %d:  ",NAME_OF_TOOL,line_no);
   vfprintf(stderr,string,ap);
   no_warnings++;
}
/*****************************************************************************

			i n f o

	 Prints a uniform informational message giving the name of the tool.

Input:	string = the information message
	... = arguments for the information string

*****************************************************************************/
void info(char *string, ...)
{
   va_list ap;
   va_start(ap,string);
   fprintf(stderr,"%s:  ",NAME_OF_TOOL);
   vfprintf(stderr,string,ap);
}
/*****************************************************************************

			p u t _ a r c h

	 Put the given string onto the architecture output.

Input:	string = the string to put.

*****************************************************************************/
void put_arch(char *string)
{
   fprintf(arch_file,"%s",string);
}
/*****************************************************************************

			p u t _ s t m t

	 Put the given string onto the statement output.

Input:	string = the string to put.

*****************************************************************************/
void put_stmt(char *string)
{
   fprintf(stmt_file,"%s",string);
}
/*****************************************************************************

			p u t _ o u t

	 Put the given string onto the "output" output.

Input:	string = the string to put.

*****************************************************************************/
void put_out(char *string)
{
   fprintf(out_file,"%s",string);
}
/*****************************************************************************

			p u t _ t a i l

	 Put the given string onto the "tail" temporary file.

Input:	string = the string to put.

*****************************************************************************/
void put_tail(char *string)
{
   fprintf(tail_file,"%s",string);
}
/************************************************************************

			p r i n t _ c o m m e n t s

      Prints out the given comment on the given file.  Space taken by the comment
is deallocated.

Input:	ofile = the file to write the comment on
	comment = the comment to write

************************************************************************/
void print_comments(FILE *ofile,comment_type *comment)
{
   comment_type *cur_comment,*old_comment;

   for (cur_comment=comment;cur_comment!=NULL;) {
      fprintf(ofile,"%s",cur_comment->string);
      free(cur_comment->string);
      old_comment=cur_comment;
      cur_comment=cur_comment->next;
      free(old_comment);
   }
}
/*****************************************************************************

			t p u t

      Puts the given string onto the appropriate temporary files according
to token_print and tp_other.

Input:	string = the string to put

*****************************************************************************/
void tput(char *string)
{
   switch (token_print) {
      case PUT_OUT: put_out(string);
		    break;
      case PUT_STMT: put_stmt(string);
		     break;
      case PUT_ARCH: put_arch(string);
		     break;
      case PUT_TAIL: put_tail(string);
		     break;
   }
   switch (tp_other) {
      case PUT_OUT: put_out(string);
		    break;
      case PUT_STMT: put_stmt(string);
		     break;
      case PUT_ARCH: put_arch(string);
		     break;
      case PUT_TAIL: put_tail(string);
		     break;
   }
}
/*****************************************************************************

			t p r i n t

	 Puts the given string onto the appropriate output.  token_print
   specifies the main output while tp_other specifies a secondary output.

Input:	string = the string to put

*****************************************************************************/
void tprint(char *string,comment_type *comment)
{
   comment_type *cur_comm,*old_comm;

   /* save string for future uses */
   concat(string);
   /* print out any comments associated with token */
   for (cur_comm=comment;cur_comm!=NULL;) {
      if (!ignore_comments)
         /* can only put comments into stmt file */
         switch (token_print) {
            case PUT_OUT: put_out(cur_comm->string);
                          break;
            case PUT_STMT: put_stmt(cur_comm->string);
                           break;
            case PUT_ARCH: put_arch(cur_comm->string);
                           break;
            case PUT_TAIL: put_tail(cur_comm->string);
                           break;
         }
      free(cur_comm->string);
      old_comm=cur_comm;
      cur_comm=cur_comm->next;
      free(old_comm);
   }
   /* print out token */
   tput(string);
}
/*****************************************************************************

			t p

	 The same as tprint.  Smaller so that the yacc grammar is more compact.

*****************************************************************************/
void tp(char *string,comment_type *comment)
{
   tprint(string,comment);
}
/*****************************************************************************

			v i t o _ c h e c k

      Checks if an implicit style block is currently being processed.  If so, then
   this portion of the grammar is unacceptable to VITO.

input:	statement = the statement not supported

*****************************************************************************/
void vito_check(char *statement)
{
   if (ai_process)
      error("VITO does not yet support the '%s' statement in implicit style code.\n",statement);
}
/************************************************************************

			i p u s h

      Pushes the given item onto the top of a dynamic integer stack.
The stack pointer is adjusted to point to the new top.

Input:	stack = address of the stack pointer
	item = the item to push

************************************************************************/
void ipush(dynamic_istack_type **stack,int item)
{
   dynamic_istack_type *new;
   new = (dynamic_istack_type *)malloc(sizeof(dynamic_istack_type));
   if (new==NULL) {
      error("Malloc returned NULL.  Ran out of memory.\n");
      exit(EXIT_FAILURE);
   }
   new->next = *stack;
   new->item = item;
   *stack = new;
}
/************************************************************************

			i p u l l

      Pulls an element off of a dynamic integer stack.  It returns the
element pulled off.  The stack pointer is modified to point to the new
top.

Input:	stack = the address of the stack pointer element
Output:	ipull = the item on top of the stack

************************************************************************/
int ipull(dynamic_istack_type **stack)
{
   dynamic_istack_type *old;
   int item;

   if (*stack==NULL) {
      error("Trying to pull from empty istack.\n");
      exit(EXIT_FAILURE);
   }
   old = *stack;
   item = old->item;
   *stack = (*stack)->next;
   free(old);
   return item;
}
/************************************************************************

			c a s e _ p u s h

      Pushes a new item onto the case stack.  The new entry is initialized
to the given expression and its counters cleared.

Input:	stack = address of stack pointer
	expression = the expression to push

************************************************************************/
void case_push(case_stack_type **stack,int expression)
{
   case_stack_type *new;
   new = (case_stack_type *)malloc(sizeof(case_stack_type));
   if (new==NULL) {
      error("Malloc returned NULL.  Ran out of memory.\n");
      exit(EXIT_FAILURE);
   }
   new->expression = expression;
   new->found_default = FALSE;
   new->no_items=0;
   new->next = *stack;
   *stack = new;
}
/************************************************************************

			c a s e _ p u l l

      Pulls an element off of the case stack.
The stack pointer is modified to point to the new
top.

Input:	stack = the address of the stack pointer element

************************************************************************/
void case_pull(case_stack_type **stack)
{
   case_stack_type *old;

   if (*stack==NULL) {
      error("Trying to pull from empty case stack.\n");
      exit(EXIT_FAILURE);
   }
   old = *stack;
   *stack = (*stack)->next;
   free(old);
}
/************************************************************************

			c a s e _ t o p

      Returns the value on the top of the case stack.  This does
not modify the stack.

Input:	stack = the stack to examine
Output:	case_top = pointer to the top of stack

************************************************************************/
case_stack_type *case_top(case_stack_type *stack)
{
   if (stack==NULL) {
      error("Trying to access empty istack.\n");
      exit(EXIT_FAILURE);
   }
   return stack;
}
/************************************************************************

			n e w _ t o k e n _ p r i n t

      Pushes the current token_print value and sets it to the given value.

Input:	new = the new token_print value

************************************************************************/
void new_token_print(put_type new)
{
   ipush(&token_print_stack,token_print);
   token_print = new;
}
/************************************************************************

			o l d _ t o k e n _ p r i n t

      Pulls the old token_print value off of the stack.

************************************************************************/
void old_token_print(void)
{
   token_print = ipull(&token_print_stack);
}
void append_library(const char *library)
{
   if (library==NULL || library[0]=='\0')
      return;
   library_paths = (char*)realloc(library_paths,strlen(library_paths)+strlen(library)+2);
   if (library_paths==NULL){
      error("Malloc ran out of room.  No more memory.\n");
      exit(EXIT_FAILURE);
   }
   if (library_paths[0]!='\0')
      strcat(library_paths," ");
   strcat(library_paths,library);
}
/************************************************************************

			a r g _ p r o c e s s e d

      Determines if arguments are processed multiple times.

Input:	arg_processed = TRUE if argument already processed
	arg = the argument character
Output:	arg_processed = always set to TRUE

************************************************************************/
void arg_processing(int *arg_processed,char arg)
{
   if (*arg_processed)
      warning("Command line argument '-%c' occurs multiple times.\n",arg);
   *arg_processed=TRUE;
}
/*****************************************************************************

			i n i t i a l i z e
		     
      Initializes all the initial settings for our tool.

*****************************************************************************/
void initialize(int argc,char *argv[])
{
   int cur_arg;
   int oarg_processed,	/* indicates if the command line argument has ever been processed */
       parg_processed,
       varg_processed,
       targ_processed,
       Parg_processed,
       darg_processed;
   int cur_ID;

   last_macro_pp = 0;
   library_paths=(char*)malloc(1);
   if (library_paths==NULL) {
      error("Malloc failed.  Ran out of memory.\n");
      exit(EXIT_FAILURE);
   }
   library_paths[0]='\0';
   token_print_stack = NULL;
   case_stack = NULL;
   current_comment = NULL;
   current_comment_end = NULL;
   current_module = -1;
   preprocess_mode = FALSE;
   verbose_mode = FALSE;
   debug_mode = FALSE;
   table_mode = FALSE;
   line_no = 1;
   no_modules = 0;
   no_errors = 0;
   no_warnings = 0;
   last_ID=0;
   rcfile_name = RCFILE;
   outfile_name = rc_string[CFFILE];
   /* process the command line arguments */
   last_source_file = 0;
   oarg_processed=FALSE;
   parg_processed=FALSE;
   varg_processed=FALSE;
   targ_processed=FALSE;
   Parg_processed=FALSE;
   darg_processed=FALSE;
   for (cur_arg=1;cur_arg<argc;cur_arg++) {
      if (strcmp(argv[cur_arg],"-o")==0) {
	 /* output file name */
	 arg_processing(&oarg_processed,'o');
	 cur_arg++;
	 if (cur_arg<argc)
	    outfile_name = argv[cur_arg];
	 else {
	    error("No file name given for -o argument.\n");
	    exit(EXIT_FAILURE);
	 }
      }
      else if (strcmp(argv[cur_arg],"-p")==0) {
	 /* prefix file name */
	 arg_processing(&parg_processed,'p');
	 cur_arg++;
	 if (cur_arg<argc)
	    rcfile_name = argv[cur_arg];
	 else {
	    error("No file name given for -p argument.\n");
	    exit(EXIT_FAILURE);
	 }
      }
      else if (strcmp(argv[cur_arg],"-l")==0) {
	 /* library path name */
	 /* allow multiple -l arguments
	 arg_processing(&larg_processed,'l');
	 */
	 cur_arg++;
	 if (cur_arg<argc) {
	    /* concatenate argument onto library_paths */
	    append_library(argv[cur_arg]);
	 }
	 else {
	    error("No path name given for -l argument.\n");
	    exit(EXIT_FAILURE);
	 }
      }
      else if (strcmp(argv[cur_arg],"-m")==0) {
	 /* macro definition */
	 cur_arg++;
	 if (cur_arg+1<argc) {
	    /* insert macro into the preprocessor's macro table */
	    cur_ID = Insert_ID_pp(argv[cur_arg],strlen(argv[cur_arg]));
	    cur_arg++;
	    if (macro_table_pp[cur_ID].perm) 
	       warning("Multiple command line definitions of macro '%s'.\n",macro_table_pp[cur_ID].ID);
	    macro_table_pp[cur_ID].macro = argv[cur_arg];
	    macro_table_pp[cur_ID].perm = TRUE;
	 }
	 else {
	    error("No macro name or macro text given for -m argument.\n");
	    exit(EXIT_FAILURE);
	 }
      }
      else if (strcmp(argv[cur_arg],"-v")==0) {
	 /* verbose output enable */
	 arg_processing(&varg_processed,'v');
	 verbose_mode = TRUE;
      }
      else if (strcmp(argv[cur_arg],"-d")==0) {
	 /* debug mode enable */
	 arg_processing(&darg_processed,'d');
	 debug_mode = TRUE;
      }
      else if (strcmp(argv[cur_arg],"-t")==0) {
	 /* debug comefrom table output */
	 arg_processing(&targ_processed,'t');
	 table_mode = TRUE;
      }
      else if (strcmp(argv[cur_arg],"-P")==0) {
	 /* perform preprocessor only */
	 arg_processing(&Parg_processed,'P');
	 preprocess_mode = TRUE;
      }
      else if (strcmp(argv[cur_arg],"-h")==0 ||
               strcmp(argv[cur_arg],"-H")==0 ||
               strcmp(argv[cur_arg],"-help")==0 ||
               strcmp(argv[cur_arg],"-HELP")==0 ||
               strcmp(argv[cur_arg],"-?")==0) {
	 /* Display help information */
	 printf("         %s\n",VERSION);
	 printf("Usage: vito [options] [file ...]\n");
	 printf("  -d               create debug files\n"); 
	 printf("  -h               display help message (also -H -help -HELP -?)\n");
	 printf("  -l <library>     specify an `include library path\n");
	 printf("  -m <name> <text> create a macro\n");
	 printf("  -o <file>        specify output file name\n");
	 printf("  -p <file>        use the named prefix file instead of .vitorc\n");
	 printf("  -P               perform only the prepocessing step\n");
	 printf("  -t               output internal comefrom tables\n");
	 printf("  -v               verbose output mode\n");
	 printf("All other arguments are taken as file names to process.\n");
	 exit(EXIT_SUCCESS);
      }
      else {
	 /* assumed to be a source file name. */
	 if (last_source_file >= MAX_NO_FILES) {
	    error("Too many source files.  Increase MAX_NO_FILES.\n");
	    exit(EXIT_FAILURE);
	 }
	 source_files[last_source_file] = argv[cur_arg];
	 last_source_file++;
      }
   }
   read_rc_file();
   rc_string[CFFILE] = outfile_name;
   append_library(rc_string[LIBRARYPATHS]);

   /* Create temporary files for internal use. */
   stmt_file = tmpfile();
   out_file = tmpfile();
   arch_file = tmpfile();
   tail_file = tmpfile();
   if (stmt_file==NULL || out_file==NULL || arch_file==NULL || tail_file==NULL) {
      error("Cannot create temporary file.\n");
      exit(EXIT_FAILURE);
   }

   if (verbose_mode)
      info(" %s\n",VERSION);
   initial_last_ID = last_ID;
}
/************************************************************************

			i n i t _ f o r _ m o d u l e

      Initialization processing for the beginning of a module in YACC.

************************************************************************/
void init_for_module(void)
{
   ts_clear();
   ai_process = FALSE;
   initial_begin = FALSE;
   always_begin = FALSE;
   do_new_wire = FALSE;
   duplicate_regs = TRUE;
   ignore_comments=FALSE;
   token_print=PUT_STMT;
   tp_other = PUT_NONE;
   /* clear out ID table */
   remove_ID(initial_last_ID);
}
/************************************************************************

			p o s t _ f o r _ m o d u l e

      Post processing when the end of a module is encountered in YACC.

************************************************************************/
void post_for_module(void)
{
   if (no_modules >= MAX_NO_MODULES) {
      error("Too many modules.  Increase MAX_NO_MODULES\n");
      exit(EXIT_FAILURE);
   }
   module_info[no_modules].implicit = implicit_module;
   no_modules++;
   put_out("\n");
   put_out(MODULE_SEPARATOR);
   put_arch("\n");
   put_arch(MODULE_SEPARATOR);
   put_stmt("\n");
   put_stmt(MODULE_SEPARATOR);
   put_tail("\n");
   put_tail(MODULE_SEPARATOR);
}
/*****************************************************************************

			f o u n d _ t o k e n

      Time to return a token to YACC.  Process any comments found by attaching
them to yylval and resetting the comment accumulator.

*****************************************************************************/
void found_token(void)
{
   yylval.comment = current_comment;
   current_comment=NULL;
   current_comment_end=NULL;
}
/*****************************************************************************

			c o m m e n t

	 Puts the given comment into the appropriate place.

Input:	string = the comment to put

*****************************************************************************/
void comment(char *string)
{
   comment_type *cur_comm,*new_comm;

   if (!ignore_comments) {
      new_comm = (comment_type *)malloc(sizeof(comment_type));
      if (new_comm==NULL) {
	 error("Ran out of memory.\n");
	 exit(EXIT_FAILURE);
      }
      new_comm->string = (char *)malloc(strlen(string)+1);
      if (new_comm->string ==NULL) {
	 error("Ran out of memory.\n");
	 exit(EXIT_FAILURE);
      }
      strcpy(new_comm->string,string);
      new_comm->next=NULL;
      if (current_comment==NULL)
	 current_comment=new_comm;
      else
	 current_comment_end->next = new_comm;
      current_comment_end=new_comm;
   }
}
/*****************************************************************************

			y y e r r o r

      The error routine for YACC.  If called, it means YACC was unable to
   successfully parse the input code.  It returns so YACC can perform
   error catch up.

Input:	error_string = the error string from YACC--not used

*****************************************************************************/
int yyerror(char *error_string)
{
   error("Syntax error on or about '%s'.\n",yytext);
}
/*****************************************************************************

			p o s t _ p a s s

	 Performs any post processing of the parsing results.
    Currently it handles:
     1.  Removing blank lines in the "out" file.
   It creates a new "out" file.

Input:	ofile = the location of the "out" file currently in use
Output:	post_pass = new out file without blank lines

*****************************************************************************/
FILE *post_pass(FILE *ofile)
{
   char string[MAXSTRING];
   FILE *out_file;
   char *ret;
   int i,isbegin;

   isbegin= -1;
   rewind(ofile);
   out_file = tmpfile();
   if (out_file==NULL) {
      error("Cannot create a temporary file.\n");
      exit(EXIT_FAILURE);
   }
   while (!feof(ofile)) {
      ret = fgets(string,MAXSTRING,ofile);
      if (feof(ofile))
	 break;
      if (ret==NULL)
	 break;
      if(string[0]!='\n' && string[0]!=LINE_TERMINATOR)
	 fputs(string,out_file);
   }
   fclose(ofile);
   return out_file;
}
/************************************************************************

			c o p y _ f i l e

      Copies the contents of the source file to the destination file.
Both files should be textual.  The source is copied until EOF is encountered.

Input:	source = the source file
	dest = the destination file

************************************************************************/
void copy_file(FILE * dest,FILE * source)
{
    char s[MAXSTRING];

    while (! feof(source))
      {
        if (fgets(s,MAXSTRING,source)==NULL)
	   break;
	fprintf(dest,"%s",s);
      }
}
/************************************************************************

			i n s e r t _ c a s e _ d e f a u l t

	 Moves the default case to the end of the nested IFs and puts
   an ELSE in front of it.  The original default statement is copied to
   the end of the file and the original overwritten with \n's.  (The
   newlines are removed by a later pass).  This is overly complicated
   but I could see no other way to do this with standard C and remain
   portable.

Input:	stack = the current case stack entry

************************************************************************/
void insert_case_default(case_stack_type *stack)
{
   char done;
   char line[MAXSTRING];
   int i;
   if (ai_process && stack->found_default && token_print==PUT_OUT){
      if (stack->no_items>1)
	 tp("else\nbegin\n",NULL);
      line[0]='\0';
      done=FALSE;
      while (!done) {
	 fsetpos(out_file,&(stack->default_start));
	 if (fgets(line,MAXSTRING,out_file)==NULL)
	    error("unable to read from out_file\n");
	 /* delete old lines */
	 fsetpos(out_file,&(stack->default_start));
	 for (i=0;i<strlen(line);i++)
	    fprintf(out_file,"%c",LINE_TERMINATOR);
	 fgetpos(out_file,&(stack->default_start));
         done = strcmp(line,CASE_DEFAULT_TERMINATOR)==0;
	 fseek(out_file,0,SEEK_END);
	 if (!done)
	    fprintf(out_file,"%s",line);
      }
      if (stack->no_items>1)
	 tp("end\n",NULL);
   }
}
/************************************************************************

			s a v e _ d e b u g

	 Saves the current set of temporary files into the named debug
files.  This is inneficient since new files are created and the contents
copied.  All that is really needed is to rename the files (if only we
knew the name of the temporary files).

************************************************************************/
void save_debug(void)
{
   FILE *debug_stmt_file,*debug_out_file,*debug_arch_file,*debug_tail_file;

   if (debug_mode) {
      debug_stmt_file = fopen(rc_string[STMTFILE],"w");
      debug_out_file = fopen(rc_string[OUTFILE],"w");
      debug_arch_file = fopen(rc_string[ARCHFILE],"w");
      debug_tail_file = fopen(rc_string[TAILFILE],"w");
      rewind(arch_file);
      rewind(stmt_file);
      rewind(out_file);
      rewind(tail_file);
      copy_file(debug_stmt_file,stmt_file);
      copy_file(debug_out_file,out_file);
      copy_file(debug_arch_file,arch_file);
      copy_file(debug_tail_file,tail_file);
   }
}
/*****************************************************************************

			m a i n

*****************************************************************************/
int main(int argc, char *argv[])
{
   int ret,i;
   FILE *output_file;

   initialize(argc,argv);

   /* Run through preprocessor */
   yyin = tmpfile();
   if (verbose_mode)
      info("Macro Preprocessing:\n");
   vitopp(yyin);

   line_no=1;
   rewind(yyin);
   if (!preprocess_mode) {
   /*
      copy_file(stdout,yyin);
      exit(EXIT_SUCCESS);
      */

   /* Parse code */
   yyparse();
   fclose(yyin);
   yyin = NULL;

   /* remove blank lines from out_file */
   out_file = post_pass(out_file);

   if (verbose_mode || no_errors>0 || no_warnings>0)
      info("Parsed %d modules with %d lines, %d errors and %d warnings.\n",no_modules,line_no-1,no_errors,no_warnings);
   if (no_errors>0) {
      save_debug();
      exit(EXIT_FAILURE);
   }
   if (verbose_mode)
      info("Converting implicit to one-hot style.\n");

   /* Convert fork-join statements */
   rewind(out_file);
   rewind(arch_file);
   fork_filter(&out_file,&arch_file,no_modules);
   }

   /* set up output file */
   if (out_filename[0]=='\0')
      output_file = stdout;
   else {
      output_file = fopen(out_filename,"wt");
      if (output_file==NULL) {
	 error("Cannot create output file '%s'.\n",out_filename);
	 exit(EXIT_FAILURE);
      }
   }
   if (preprocess_mode){
      copy_file(output_file,yyin);
      exit(EXIT_SUCCESS);
   }
   rewind(arch_file);
   rewind(stmt_file);
   rewind(out_file);
   rewind(tail_file);

   /* Perform comefrom conversion */
   for (current_module=0;current_module<no_modules;current_module++)
      come_from(output_file,module_info[current_module]);

   /* print out comments after last module */
/*   print_comments(output_file,current_comment);
  */ save_debug();
   fclose(output_file);
   fclose(stmt_file);
   fclose(out_file);
   fclose(arch_file);
   fclose(tail_file);
   if (verbose_mode || no_errors>0 || no_warnings>0)
      info("Converted %d module%s with %d lines, %d errors and %d warnings.\n",no_modules,no_modules==1?"":"s",line_no-1,no_errors,no_warnings);
   exit(EXIT_SUCCESS);
}
