/*****************************************************************************
*                                                                            *
*     VITO -- Verilog Implicit To One-hot translator                         *
*     $Id: legal.txt,v 1.7 1999/06/03 15:24:08 mwinkel Apr_19_2004_Version_1.5 $
*                                                                            *
*     This software translates implicit style Verilog into a form suitable   *
*     for most commercial synthesis tools.  The translation is achieved      *
*     by constructing a one-hot machine for the ASM described in the         *
*     implicit style Verilog.                                                *
*                                                                            *
*     This work is being performed at the University of Wyoming and SUNY     *
*     College at Brockport.  For more information:                           *
*	   www.verilog.vito.com                                              *
*          http://strawberry.uwyo.edu/~vito                                  *
*          http://cs.brockport.edu/~jshuler/vito                             *
*                                                                            *
*     Authors:                                                               *
*          James D. Shuler  jshuler@cs.brockport.edu                         *
*          Mark G. Arnold   marnold@uwyo.edu                                 *
*                                                                            *
*                         NOTICE AND DISCLAIMER                              *
*                                                                            *
* No use of this software is authorized except under this NOTICE AND         *
* DISCLAIMER.  Downloading, use, copying, modification, and/or               *
* distribution of this software in its original or modified form implies     *
* your acceptance of this NOTICE and DISCLAIMER.                             *
*                                                                            *
* Permission to use, copy, modify and distribute this software and its       *
* documentation, for any purpose and without fee is hereby granted,          *
* provided that this permission notice appears prominently in supporting     *
* documentation and must be viewed prior to use, copying, modification or    *
* distribution.  You are responsible for any modifications to the software   *
* which you make and notice that the software has been modified must be      *
* appended to this notice prior to further use, copying, modification or     *
* distribution.                                                              *
*                                                                            *
* This software is provided AS IS without warranty of any kind,              *
* including without limitation the warranties that the software is           *
* non-infringing, merchantable, or fit for a particular purpose, including   *
* high risk activities.  The entire risk as to the quality and performance   *
* of the software is born by you.  Should the software prove defective in    *
* any respect, you and not the developers, nor any parties associated with   *
* the developers, assume the entire cost of  any service and repair.         *
*                                                                            *
* This software may be subject to the Export Control Laws of the United      *
* States of America.  It is your responsibility to determine the             *
* applicable laws and regulations and comply with them.                      *
*                                                                            *
*****************************************************************************/
/*****************************************************************************
$Id: vito.h,v 1.7 1999/07/21 20:42:42 mwinkel Apr_19_2004_Version_1.5 $

	Header file for VITO programs containing global definitions and
   prototypes.

*****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*****************************************************************************

			G l o b a l   C o n s t a n t s

*****************************************************************************/
#define FALSE		0
#define TRUE		1

#define UNIX		1	/* UNIX version of VITO */
#define DOS		2	/* DOS version of VITO */
#define OPERATING_SYSTEM	UNIX	/* Which system to compile for */

#define NAME_OF_TOOL	"//VITO"  /* name of our Verilog processing tool */
#define VERSION		"VITO $State: Apr_19_2004_Version_1.5 $ "	/* version info on VITO */
#if OPERATING_SYSTEM == UNIX
   #define RCFILE	".vitorc" /* initialization file name*/
#elif OPERATING_SYSTEM == DOS
   #define RCFILE	"vito.rc" /* initialization file name*/
#endif
#define MODULE_SEPARATOR "/* VITO MODULE SEPARATOR */\n"	/* separates modules in the intermediate files */
#define CASE_DEFAULT_TERMINATOR "/* CASE DEFAULT TERMINATOR */\n" /* terminates the end of the default case item */
#define RC_COMMENT	'%'	/* comment character for RC file */
#if OPERATING_SYSTEM == UNIX
   #define LINE_TERMINATOR	'\n'	/* indicates end of line for case_insert and filer */
#elif OPERATING_SYSTEM == DOS
   #define LINE_TERMINATOR	'\r'	/* indicates end of line for case_insert and filer */
#endif
#define MAXSTRING	4096	/* maximum length of a string */
#define MAX_TOKEN_STRING MAXSTRING	/* maximum length of the token string read in so far */
#define MAXLEXSTRING	1024	/* maximum number of characters for lex to store before being printed */
#define MAXID		1024	/* maximum number of distinct identifiers and numbers allowed */
#define MAXMACRO	1024	/* Maximum number of macro definitions allowed */
#define MAX_MACRO_DEPTH	1000	/* Maximum number of recursive macros, includes, or ifdefs */
#define MAXNOERRORS	20	/* maximum number of errors allowed before aborting */
#define MAX_NO_FILES	10	/* maximum number of source files */
#define MAX_NO_MODULES	100	/* maximum number of modules in code */

#define MAXRC		18	/* number of lines in RC file */
/*  Lines in the RCFILE  */
#define OUTFILE		0	/* name of file for implicit block statements */
#define STMTFILE 	1	/* name of file for non-implicit block statements */
#define ARCHFILE 	2	/* name of file for architecture statements */
#define CFFILE		3	/* name of file for comefrom output */
#define	WIREPRE		4	/* prefix for wire names */
#define TWIREPRE	5	/* prefix for true path wire names */
#define TEMPPRE		6	/* prefix for temporary wires */
#define	JOINPRE		7	/* prefix for join regs */
#define FFOUTPUT 	8	/* flop-flop output name */
#define QUALPRE		9	/* prefix for qualifier wire */
#define RESET		10	/* name of reset signal */
#define CLOCK		11	/* name of clock signal */
#define ATEVENT		12	/* string for @() events */
#define NEWWIREPRE	13	/* name of new wire prefix */
#define RESETEDGE	14	/* event for causing reset? */
#define RESETCOND	15	/* condition for performing reset */
#define TAILFILE 	16	/* name of file for tail statements */
#define LIBRARYPATHS	17	/* pathnames for library files */

/*  Definitions for YACC */
typedef struct comment_typep {
	    char *string;
	    struct comment_typep *next;
	 } comment_type;
typedef struct dynamic_istackp {
	    int item;
	    struct dynamic_istackp *next;
	 } dynamic_istack_type;
typedef struct case_stackp {
	    int expression;
	    char found_default;
	    int no_items;
	    fpos_t default_start;
	    struct case_stackp *next;
	 } case_stack_type;
typedef  struct{
	    int identifier;
	    comment_type *comment;
	 } my_yystype;
#define YYSTYPE my_yystype

/*  Places to put text during parsing  */
typedef enum {PUT_NONE,		/* Do not put text anywhere */
	      PUT_STMT,		/* Append text to statement file */
	      PUT_ARCH,		/* Append text to architecture file */
	      PUT_TAIL,		/* Append text to tail file */
	      PUT_OUT		/* Append text to out file */
	      } put_type;
/* information about modules */
typedef struct {
   int implicit;
} module_info_type;
/* macro table */
typedef struct {
   char *ID;		/* the identifier string */
   char *macro;		/* the macro string (NULL if none) */
   int perm;		/* TRUE if macro should not be redefinable */
} macro_table_type;

/*****************************************************************************

			E x t e r n    V a r i a b l e s

*****************************************************************************/
extern int ignore_comments;
extern int ai_process;
extern int initial_begin,always_begin;
extern char *yytext;
extern int yyleng;
extern YYSTYPE yylval;
extern FILE *yyin;
extern FILE *yyout;
extern put_type token_print,tp_other;
extern char *rc_string[MAXRC];
extern char token_string[MAX_TOKEN_STRING];
extern int ts_length;
extern int do_new_wire;
extern int duplicate_regs;
extern int line_no;
extern int no_errors;
extern int reset_ID,clock_ID;
extern int reset_in_port,clock_in_port;
extern int last_source_file;
extern char *source_files[];
extern int verbose_mode;
extern int debug_mode;
extern int table_mode;
extern FILE *out_file,*stmt_file,*arch_file,*tail_file;
extern char *rc_string[MAXRC];	/* Array of initialization strings */
extern int current_module;
extern int implicit_module;
extern case_stack_type *case_stack;
extern char *library_paths;
extern macro_table_type macro_table_pp[];
extern int last_macro_pp;

/*****************************************************************************

			E x t e r n a l    F u n c t i o n s

*****************************************************************************/
void tprint(char *string,comment_type *comment);
void tp(char *string,comment_type *comment);
void put_stmt(char *string);
void put_stmtnl(char *string);
void put_out(char *string);
void put_outnl(char *string);
void comment(char *string);
char *Get_ID(int location);
int Insert_ID(char *ID,int length);
void error(char *string, ...);
void warning(char *string, ...);
void info(char *string, ...);
void vito_check(char *);
int yywrap(void);
void copy_file( FILE * outf, FILE * f);
void copy_file_module( FILE * outf, FILE * f);
void case_push(case_stack_type **stack,int expression);
void case_pull(case_stack_type **stack);
case_stack_type *case_top(case_stack_type *stack);
void insert_case_default(case_stack_type *stack);
void ipush(dynamic_istack_type **,int);
int ipull(dynamic_istack_type **);
void new_token_print(put_type new);
void old_token_print(void);
void append_library(const char *library);
int Insert_ID_pp(char *ID,int length);
