/*****************************************************************************
*                                                                            *
*     VITO -- Verilog Implicit To One-hot translator                         *
*     $Id: legal.txt,v 1.7 1999/06/03 15:24:08 mwinkel Apr_19_2004_Version_1.5 $
*                                                                            *
*     This software translates implicit style Verilog into a form suitable   *
*     for most commercial synthesis tools.  The translation is achieved      *
*     by constructing a one-hot machine for the ASM described in the         *
*     implicit style Verilog.                                                *
*                                                                            *
*     This work is being performed at the University of Wyoming and SUNY     *
*     College at Brockport.  For more information:                           *
*	   www.verilog.vito.com                                              *
*          http://strawberry.uwyo.edu/~vito                                  *
*          http://cs.brockport.edu/~jshuler/vito                             *
*                                                                            *
*     Authors:                                                               *
*          James D. Shuler  jshuler@cs.brockport.edu                         *
*          Mark G. Arnold   marnold@uwyo.edu                                 *
*                                                                            *
*                         NOTICE AND DISCLAIMER                              *
*                                                                            *
* No use of this software is authorized except under this NOTICE AND         *
* DISCLAIMER.  Downloading, use, copying, modification, and/or               *
* distribution of this software in its original or modified form implies     *
* your acceptance of this NOTICE and DISCLAIMER.                             *
*                                                                            *
* Permission to use, copy, modify and distribute this software and its       *
* documentation, for any purpose and without fee is hereby granted,          *
* provided that this permission notice appears prominently in supporting     *
* documentation and must be viewed prior to use, copying, modification or    *
* distribution.  You are responsible for any modifications to the software   *
* which you make and notice that the software has been modified must be      *
* appended to this notice prior to further use, copying, modification or     *
* distribution.                                                              *
*                                                                            *
* This software is provided AS IS without warranty of any kind,              *
* including without limitation the warranties that the software is           *
* non-infringing, merchantable, or fit for a particular purpose, including   *
* high risk activities.  The entire risk as to the quality and performance   *
* of the software is born by you.  Should the software prove defective in    *
* any respect, you and not the developers, nor any parties associated with   *
* the developers, assume the entire cost of  any service and repair.         *
*                                                                            *
* This software may be subject to the Export Control Laws of the United      *
* States of America.  It is your responsibility to determine the             *
* applicable laws and regulations and comply with them.                      *
*                                                                            *
*****************************************************************************/
/*****************************************************************************
$Id: vitocf.c,v 1.5 1999/06/03 15:54:04 mwinkel Apr_19_2004_Version_1.5 $

This module implements the comefrom pass of VITO.  It translates the implicit
style always blocks into one-hot style.

*****************************************************************************/
#include "vito.h"

#define MAXID_M 30
#define MAXASSIGN 500
#define MAXSTRING_M 256
#define MAXSP 100
#define MAXLINE 999
#define TIF 1
#define TWHILE 2
#define TFOREVER 3
#define TELSE 4
#define TBEGIN 5

typedef struct
  {
    int sp;
    int s[MAXSP];
    int skind[MAXSP];
    char sname[MAXSP] [MAXID_M];
    int sbexit[MAXSP];
  }stacktype;

typedef struct
  {
    int a[MAXLINE];
    int b[MAXLINE];
    unsigned char ff[MAXLINE];
    unsigned char panos[MAXLINE]; 
  }comefromtype;

typedef struct
  {
    int num;
    char lval[MAXASSIGN] [MAXID_M];
    char expr[MAXASSIGN] [MAXSTRING_M];
    int lineno[MAXASSIGN];
  }assigntype;

#define ctrlstrc_filename	rc_string[OUTFILE]
#define unchstmt_filename	rc_string[STMTFILE]
#define arch_filename		rc_string[ARCHFILE]
#define out_filename		rc_string[CFFILE]
#define s_name			rc_string[WIREPRE]
#define st_name			rc_string[TWIREPRE]
#define tmp_name		rc_string[TEMPPRE]
#define ff_modname		rc_string[FFMODULE]
#define ff_name			rc_string[FFOUTPUT]
#define qual_name		rc_string[QUALPRE]
#define reset_name		rc_string[RESET]
#define sysclk_name		rc_string[CLOCK]
#define edgeevent_name		rc_string[ATEVENT]
#define newreg_name		rc_string[NEWWIREPRE]
#define edgeresetevent_name	rc_string[RESETEDGE]
#define resetcond_name		rc_string[RESETCOND]


void push(
	  stacktype * stack,
	  int val,
	  int kind,
	  char * name
	 )
  {
    (*stack).sp++;
    if ((*stack).sp == MAXSP)
      {
	error("Comefrom: Nested too deep.  Increase MAXSP.\n");
        exit(EXIT_FAILURE);
      }
    (*stack).s[(*stack).sp] = val;
    (*stack).skind[(*stack).sp] = kind;
    strcpy((*stack).sname[(*stack).sp], name);
    (*stack).sbexit[(*stack).sp] = 0;
  }



void pop(
          stacktype * stack,
          int * top,
          int * topkind,
          char * topname,
          int * topexit
             )
  {
    if ((*stack).sp == 0)
      {
	error("Comefrom:  Too many ends.\n");
        exit(EXIT_FAILURE);
      }
    *top = (*stack).s[(*stack).sp];
    *topkind = (*stack).skind[(*stack).sp];
    strcpy(topname, (*stack).sname[(*stack).sp]);
    *topexit = (*stack).sbexit[(*stack).sp];
    (*stack).sp--;
  }



void createstack(
		 stacktype * stack
                )
   {
     (*stack).sp = 0;
   }



void processbegin(
                   FILE * f,
                   int * n,
		   comefromtype * comefrom,
                   stacktype * stack,
                   int token
                 )
  {
    char s[MAXSTRING_M];
    char * poscolon;
    char blockname[MAXID_M];

    if ((token == TIF) || (token == TWHILE))
      (*comefrom).a[*n+1] = -*n;
    else
      (*comefrom).a[*n+1] = *n;
    (*comefrom).b[*n+1] = 0;

    fgets(s,MAXSTRING_M,f);
    (*n)++;
    if (strstr(s,"begin") != s)
      {
	error("Comefrom: Lacks begin-end pair enclosing statements\n");
	/*
        exit(EXIT_FAILURE);
	*/
      }
    poscolon = strchr(s,':');
    if (poscolon != NULL)
      strcpy(blockname, poscolon+1);
    else
      strcpy(blockname,"");
    push(stack,*n-1,token,blockname);
    (*comefrom).a[*n+1] = *n;
    (*comefrom).b[*n+1] = 0;
  }



void initcomefrom(
		comefromtype * comefrom
		  )
  {
    int i;

    for (i = 0; i < MAXLINE; i++)
      {
	(*comefrom).a[i] = MAXLINE;
	(*comefrom).b[i] = MAXLINE;
        (*comefrom).ff[i] = 0;       //bstmp and cuda
        (*comefrom).panos[i] = 0;    //panos 
      }
  }



void pass2(
	    comefromtype * comefrom,
            int * n
          )
  {
    FILE * f;
    char s[MAXSTRING_M];
    int kind;
    int oldn;
    stacktype stack;
    int i;
    char blockname[MAXSTRING_M];
    int blockexit;
    char * poscolon;
    char disablelabel[MAXSTRING_M];
    int matchsp;
    int searchsp;

    createstack(&stack);
    initcomefrom(comefrom);
    f = out_file;
    if (feof(f)) error("Empty file\n");

    *n = 0;
    while (!feof(f))
      {
	fgets(s,MAXSTRING_M,f);
	if (feof(f) || strcmp(MODULE_SEPARATOR,s)==0) break;
	(*n)++;
	if (*n == MAXLINE)
          {
            error("Comefrom:  Too many behavioral statements\n");
            exit(EXIT_FAILURE);
          }



        if   ((strstr(s, "while(" ) == s) ||   //panos
	      (strstr(s, "if(" ) == s)    ||   //panos
	      (strstr(s,edgeevent_name) == s)) //panos
  	   (*comefrom).panos[*n] = 1; 



        if        (strstr(s, "while(" ) == s)
          processbegin(f, n, comefrom, &stack, TWHILE);
        else if   (strstr(s, "if(" ) == s)
          processbegin(f, n, comefrom, &stack, TIF);
	else if   (strstr(s, "always" ) == s)
	  {
	     (*comefrom).a[*n] = MAXLINE;
	     processbegin(f, n, comefrom, &stack, TFOREVER);
	  }
	else if   (strstr(s, "forever" ) == s)
	  processbegin(f, n, comefrom, &stack, TFOREVER);
        else if   (strstr(s, "else") == s)
          processbegin(f, n, comefrom, &stack, TELSE);
        else if   (strstr(s, "begin") == s)
          {
            poscolon = strchr(s,':');
	    if (poscolon != NULL)
              strcpy(blockname,poscolon+1);
            else
	      strcpy(blockname, "");
	    push(&stack,*n-1,TBEGIN,blockname);
	    (*comefrom).a[*n+1] = *n;
	    (*comefrom).b[*n+1] = 0;
	  }
	else if   (strstr(s, "disable ") == s)
	  {
	    strncpy(disablelabel, &s[8], strlen(s)-9);
	    disablelabel[strlen(s)-10] = '\n';
	    disablelabel[strlen(s)-9] = '\0';
	    matchsp = -1;
	    for (searchsp = stack.sp; searchsp >=0; searchsp--)
	      {
		if (!strcmp(stack.sname[searchsp],disablelabel))
		  matchsp = searchsp;
	      }
	    if (matchsp == -1)
	      {
		error("Comefrom:  bad label\n");
		exit(EXIT_FAILURE);
	      }

	    (*comefrom).b[*n] = stack.sbexit[matchsp];
	    stack.sbexit[matchsp] = *n;
	    (*comefrom).a[*n+1] = 0;
	    (*comefrom).b[*n+1] = 0;
	  }
	else if   (strstr(s, "end" ) == s)
          {
            pop(&stack,&oldn,&kind,blockname,&blockexit);
	    if (kind == TWHILE)
              {
                (*comefrom).b[*n] = blockexit;
                (*comefrom).a[*n+1] = oldn;
                (*comefrom).b[oldn] = *n;
                (*comefrom).b[*n+1] = 0;
              }
            if (kind == TIF)
              {
                if (blockexit != 0)
                  {
                    error("Comefrom:  Illegal disable from if\n");
                    exit(EXIT_FAILURE);
                  }
                (*comefrom).b[*n] = oldn;
                (*comefrom).a[*n+1] = *n;
                (*comefrom).b[*n+1] = 0;
              }
            if (kind == TFOREVER)
              {
                (*comefrom).b[*n] = blockexit;
                (*comefrom).a[*n+1] = 0;
                (*comefrom).b[oldn] = *n;
                (*comefrom).b[*n+1] = 0;
              }
            if (kind == TELSE)
              {
                if (blockexit != 0)
                  {
                    error("Comefrom:  Illegal disable from else\n");
                    exit(EXIT_FAILURE);
                  }
                (*comefrom).b[*n] = oldn-1;
                (*comefrom).a[oldn] = (*comefrom).b[oldn-1];
                (*comefrom).b[oldn-1] = 0;
                (*comefrom).a[*n+1] = *n;
                (*comefrom).b[*n+1] = 0;
              }
            if (kind == TBEGIN)
              {
                (*comefrom).b[*n] = blockexit;
                (*comefrom).a[*n+1] = *n;
                (*comefrom).b[*n+1] = 0;
              }
          }
        else
          {
	    (*comefrom).a[*n+1] = *n;
	    (*comefrom).b[*n+1] = 0;
          }
      }
  }

void declarestmt(
                   FILE * outf,
                   comefromtype * comefrom,  //panos
                   int n
                )
  {
    int i;

    if (n==0) return; /* by JDS to handle non-implicit style code */

    //*$*$* fprintf(outf,"wire ");
    for (i = 1; i <= n-1; i++)
      {
        if (((*comefrom).panos[i] == 0) &&      //panos
            ((*comefrom).b[i] == 0)     )       //panos
                 fprintf(outf,"'");
        fprintf(outf,"%s%d var bit\n",s_name,i);
        //if (i%10 == 0)
        //  {
        //    fprintf(outf,"\n");
        //    fprintf(outf,"        ");
        //  }
      }          
    if (((*comefrom).panos[n] == 0) &&      //panos 1/4/10
        ((*comefrom).b[n] == 0)     )       //panos 1/4/10
             fprintf(outf,"'");
    fprintf(outf,"%s%d var bit\n",s_name,n);
    fprintf(outf,"%s%d var bit\n",s_name,MAXLINE);
  }




void genwire(
              FILE * outf,
              comefromtype * comefrom,   //panos
              int wireno
            )

  {
    if      (wireno == 0)
      fprintf(outf, "0");
    else if (wireno == MAXLINE)               //6/8/10 fixes recur bug on ellsey
      fprintf(outf, "%s%d", s_name, MAXLINE);
    else if (wireno < 0)
      fprintf(outf, "%s%d", st_name, -wireno);
    else if (((*comefrom).b[wireno]!=0)||((*comefrom).panos[wireno]!=0))  
      fprintf(outf, "%s%d", s_name, wireno);
    else                                                 //panos
      genwire(outf,comefrom,(*comefrom).a[wireno]);               //panos
  }



void generateorAlone(
                 FILE * outf,
                 int n,
                 comefromtype * comefrom
               )

  {
    genwire(outf, comefrom, (*comefrom).a[n]);         //panos
    if ((*comefrom).b[n] != 0)
      {
        fprintf(outf,"|");
	genwire(outf, comefrom, (*comefrom).b[n]);     //panos
      }
  }



void generateor(
                 FILE * outf,
                 int n,
                 comefromtype * comefrom
               )

  {
    fprintf(outf, " = ");
    generateorAlone(outf, n, comefrom);
    fprintf(outf, "\n");
  }




//void generateor(
//                 FILE * outf,
//                 int n,
//                 comefromtype * comefrom
//               )
//
//  {
//    fprintf(outf, " = ");
//    genwire(outf, comefrom, (*comefrom).a[n]);       //panos
//    if ((*comefrom).b[n] != 0)
//      {
//        fprintf(outf,"|");
// 	genwire(outf, comefrom, (*comefrom).b[n]);   //panos
//      }
//    fprintf(outf, "\n");
//  }


void behavflipflop(
                   FILE * outf,
                   int n,
                   int datin
                  )
  {
    //fprintf(outf,"  reg %s%d;\n",ff_name,n);
    //fprintf(outf,"  always %s\n",edgeresetevent_name);
    //fprintf(outf,"    if (%s)\n",resetcond_name);
    //fprintf(outf,"      %s%d = 0;\n",ff_name,n);
    //fprintf(outf,"    else\n");
    fprintf(outf,"%s%d = ",s_name,n);  //panos ff
    if (datin == 1)
      fprintf(outf,"1\n");
    else
      fprintf(outf,"%s%d\n",tmp_name,n);
  }




void pass3(
            FILE * outf,
	    comefromtype * comefrom,
	    assigntype * ass,
            fpos_t out_file_pos
          )

  {
    FILE * f;
    char s[MAXSTRING_M];
    int n;
    char * qualifier;
    char * posnb;
    char * posedge;
    int i; //for bstmp

    f = out_file;
    fsetpos(f,&out_file_pos);



    fprintf(outf,"oneHot:\n");
    fprintf(outf,"oneHotCnt var nib\n");
    fprintf(outf,"for oneHotCnt=0 to 2\n");
    (*ass).num = 0;
    n = 0;
    while (!feof(f))
      {
        fgets(s,MAXSTRING_M,f);
	if (feof(f) || strcmp(MODULE_SEPARATOR,s)==0) break;
//        if (n==0) {
//????????           /* output flip-flop 999 to start machine */
//       }
        n++;
        if      ((strstr(s, "while(" ) == s) ||
		 (strstr(s, "if(" ) == s))
          {
            qualifier = strchr(s,'(')+1;
	    *strchr(qualifier,'\n') = '\0';
//            fprintf(outf,"%s%d var bit\n",tmp_name,n);
//            fprintf(outf,"%s%d",tmp_name,n);
//	    generateor(outf, n, comefrom);
//            fprintf(outf,"%s%d var bit\n",qual_name,n);
//            fprintf(outf,"%s%d = (%s\n",
//	      qual_name,n,qualifier);
//            fprintf(outf,"%s%d = %s%d&(~%s%d)\n",
//	      s_name,n,tmp_name,n,qual_name,n);
//            fprintf(outf,"%s%d var bit\n",
//              st_name,n);
//            fprintf(outf,"%s%d = %s%d&%s%d\n",
//	      st_name,n,tmp_name,n,qual_name,n);


            fprintf(outf,"%s%d = (", s_name,n);
            generateorAlone(outf, n, comefrom);
            fprintf(outf,")&(~(%s)\n",qualifier);
            fprintf(outf,"%s%d var bit\n", st_name,n);
            fprintf(outf,"%s%d = (",st_name,n);
            generateorAlone(outf, n, comefrom);
            fprintf(outf,")&(%s\n",qualifier);




	  }
	else if (strstr(s,edgeevent_name) == s)
	  {
            fprintf(outf,"%s%d var bit\n",tmp_name,n);
            fprintf(outf,"%s%d ",tmp_name,n);
            generateor(outf, n, comefrom);

            (*comefrom).ff[n] = 1;  //added for bstmp
            //printf(outf,"%s%d var bit\n",ff_name,n); //panos ff

            //fprintf(outf,"%s%d = %s%d\n", s_name, n, ff_name, n);  //panos ff
	  }
	else
	  {
            if (((*comefrom).panos[n] == 0) &&      //panos
                ((*comefrom).b[n] == 0)     )       //panos
              fprintf(outf,"'");
            fprintf(outf,"%s%d ",s_name,n);
	    generateor(outf, n, comefrom);
	    posnb = strstr(s,"<=");
	    if (posnb != NULL)
	      {
		posedge = strstr(s,edgeevent_name);
		if (posedge != posnb+2)
		  {
		    error("Comefrom:  Bad <=\n");
		    exit(EXIT_FAILURE);
		  }
		(*ass).num++;
                if ((*ass).num >= MAXASSIGN)
                  {
                    error("Comefrom:  too many <=\n");
                    exit(EXIT_FAILURE);
		  }
		strncpy((*ass).lval[(*ass).num],s,(int)(posnb-s));
		(*ass).lval[(*ass).num] [(int)(posnb-s)] = '\0';
		strcpy((*ass).expr[(*ass).num],
                       posedge+strlen(edgeevent_name));
		(*ass).expr[(*ass).num][strlen((*ass).expr[(*ass).num])-2] = ' ';
		(*ass).expr[(*ass).num][strlen((*ass).expr[(*ass).num])-1] = '\0';
		(*ass).lineno[(*ass).num] = n;
              }
          }
      }

    fprintf(outf,"next\n");
    fprintf(outf,"%s%d=0\n",s_name,MAXLINE);
    fprintf(outf,"return\n");
    fprintf(outf,"''''end of oneHot\n\n\n");    
    fprintf(outf,"debugOneHot:\n");
    fprintf(outf,"debug %cstate %c,",'"','"');
    for (i=1; i<=MAXLINE; i++)
    {
      if ((*comefrom).ff[i])
        {
          fprintf(outf, "hex %s%d,",s_name,i);  //panos ff
        }
    }
    fprintf(outf,"hex %s%d,13\n",s_name,MAXLINE);
    fprintf(outf,"return\n");
    fprintf(outf,"''''end of debugOneHot\n\n\n");    

    fprintf(outf,"clockEvent:\n");  //section for bstmp moved for panos ff
//    for (i=1; i<=MAXLINE; i++)
//    {
//      if ((*comefrom).ff[i])
//        {
//          behavflipflop(outf, i, 0);
//        }
//    }


  }




void sortassign(
                assigntype * ass
               )

  {
    int i,j;
    char tlval[MAXID_M];
    char texpr[MAXSTRING_M];
    int tlineno;
    char oldlval[MAXID_M];

    for (i = 1; i <= (*ass).num; i++)
      {
         if (strstr((*ass).lval[i],"[")!=NULL)
           {
             strcpy(tlval,(*ass).lval[i]);
             strcpy((*ass).lval[i], " ");
             strcat((*ass).lval[i],tlval);
           }
      }
    for (i = 1; i <= (*ass).num; i++)
      {
        for (j = i; j <= (*ass).num; j++)
          {
            if (strcmp((*ass).lval[i], (*ass).lval[j]) > 0)
              {
                strcpy(tlval, (*ass).lval[i]);
                strcpy((*ass).lval[i], (*ass).lval[j]);
                strcpy((*ass).lval[j], tlval);
                strcpy(texpr, (*ass).expr[i]);
                strcpy((*ass).expr[i], (*ass).expr[j]);
                strcpy((*ass).expr[j], texpr);
                tlineno = (*ass).lineno[i];
                (*ass).lineno[i] = (*ass).lineno[j];
                (*ass).lineno[j] = tlineno;
              }
          }
      }
  }

void printoldlval(
                  FILE * outf,
                  char * oldlval
                 )

  {
    /* by JDS to solve problem with trivial module */
    if (oldlval[0]=='\0')
       return;

    fprintf(outf,"%s",oldlval);
    //fprintf(outf,"\n");
    //fprintf(outf,"  always %s\n",edgeevent_name);
    //fprintf(outf,"    %s = %s%s;\n",oldlval,newreg_name,oldlval);
    //fprintf(outf,"\n");
  }



void printarch(
               FILE * outf,
               assigntype * ass,
               comefromtype * comefrom    //panos
              )

  {
    int i,maxnum,j,oldi;
    char oldlval[MAXID_M];

    int i_find; //6/8/10
    char * pstr;

    strcpy(oldlval,(*ass).lval[1]);
    oldi = 1;
    
//    fprintf(outf,"  always %s\n",edgeevent_name);
//    fprintf(outf,"   begin\n");
    maxnum = (*ass).num;
    strcpy((*ass).lval[maxnum+1], "");
    
    for (i = 1; i<= maxnum+1; i++)
      {
	if (strcmp((*ass).lval[i], oldlval))
	  {

            for (i_find=i; i_find<=maxnum+1; i_find++) //6/8/10 start tmp_ dep
            {
              pstr = strstr((*ass).expr[i_find],oldlval);
              if(pstr)
                break;
            }
            if (pstr)      //6/8/10
            {
              fprintf(outf,"%s%s var word\n",tmp_name,oldlval);
              fprintf(outf,"%s",tmp_name);
            }                                           //end of tmp_ dep 6/8/10

            fprintf(outf,"%s=",oldlval);
            for (j=oldi; j<=i-1; j++)
	    {
                fprintf(outf,"(");                                  //panos
                genwire(outf,comefrom,(*ass).lineno[j]);            //panos
                fprintf(outf, "*((%s)-%s))+", (*ass).expr[j], oldlval); //panos
	    }
            fprintf(outf,"%s\n",oldlval);
	    strcpy(oldlval, (*ass).lval[i]);
	    oldi = i;
          }
      }

      strcpy(oldlval,(*ass).lval[1]);

      for (i = 1; i<= maxnum+1; i++)  //6/8/10 start of tmp_ depend
      {
        if (strcmp((*ass).lval[i], oldlval))
        {
          for (i_find=i; i_find<=maxnum+1; i_find++)
          {
            pstr = strstr((*ass).expr[i_find],oldlval);
            if(pstr)
            { //fprintf(outf,"found:  %s\n",pstr);     
              fprintf(outf,"%s=%s%s\n",oldlval,tmp_name,oldlval);     
              break;
            }
          }                          //6/8/10  end of tmp_ depend
          strcpy(oldlval, (*ass).lval[i]);
        }
      }                             

    //this section for bstmp moved for panos ff
    for (i=1; i<=MAXLINE; i++)
    {
      if ((*comefrom).ff[i])
        {
          behavflipflop(outf, i, 0);
        }
    }

    fprintf(outf,"return\n");
    fprintf(outf,"''''end of clockEvent\n");
  }



void oldprintarch(
               FILE * outf,
               assigntype * ass
              )

  {
    int i;
    char oldlval[MAXID_M];

    strcpy(oldlval,"");
    fprintf(outf,"\n");
    fprintf(outf,"  always %s\n",edgeevent_name);
    fprintf(outf,"   begin\n");
    for (i = 1; i<= (*ass).num; i++)
      {
	if (strcmp((*ass).lval[i], oldlval))
	  {
            if (strlen(oldlval) != 0)
              printoldlval(outf,oldlval);
            fprintf(outf,"     %s <= \n",(*ass).lval[i]);
	    strcpy(oldlval, (*ass).lval[i]);
          }
        fprintf(outf,"    %s%d ? %s :\n",s_name,(*ass).lineno[i],(*ass).expr[i]);
      }
    printoldlval(outf,oldlval);
    fprintf(outf,"   end\n");
  }





void printcomefrom(
		FILE *ofile,
		comefromtype * comefrom,
                fpos_t out_file_pos
                  )
  {
    int n,i;
    FILE * f;
    char s[MAXSTRING_M];

    /* by JDS
    f = fopen(ctrlstrc_filename,"rt");
    */
    f = out_file;
    fsetpos(f,&out_file_pos);
    /*
    rewind(f);
    */

    fprintf(ofile,"'//\n'//  Comefrom Table for module number %d:\n'//\n",current_module+1);
    n = 0;
    while (! feof(f))
      {
        fgets(s,MAXSTRING_M,f);
	if (feof(f) || strcmp(MODULE_SEPARATOR,s)==0) break;
	n++;
        fprintf(ofile,"'// %3d: %3d %3d %s",n,(*comefrom).a[n],(*comefrom).b[n],s);
      }
  }

void copy_file_module(
               FILE * outf,
               FILE * f
              )
  {
    char s[MAXSTRING_M];

    while (! feof(f))
      {
        fgets(s,MAXSTRING_M,f);
	if (feof(f) || strcmp(MODULE_SEPARATOR,s)==0) break;
	fprintf(outf,"%s",s);
      }
  }



come_from(FILE *outf, module_info_type module_info)
  {
    comefromtype comefrom;
    int numstmt;
    assigntype ass;
    FILE *atail_file;
    fpos_t out_file_pos;	/* starting position of out_file for the current module */
    fpos_t tail_file_pos;	/* starting position of tail_file for the current module */

    fgetpos(out_file,&out_file_pos);

    copy_file_module(outf,stmt_file);
    if (module_info.implicit) {
       /*copy_file_module(outf,arch_file);*/
       pass2(&comefrom, &numstmt);
       declarestmt(outf, &comefrom, numstmt);      //panos
       pass3(outf, &comefrom, &ass, out_file_pos);
       sortassign(&ass);  /*qsort??*/
       printarch(outf,&ass,&comefrom);             //panos
    }
    copy_file_module(outf,tail_file);

    if (module_info.implicit) {
       if (debug_mode) {
          fgetpos(tail_file,&tail_file_pos);
          fseek(tail_file,0,SEEK_END);
          printcomefrom(tail_file,&comefrom,out_file_pos);
          fsetpos(tail_file,&tail_file_pos);
       }
   
       if (table_mode)
          printcomefrom(outf,&comefrom,out_file_pos);
    }
}
