/*****************************************************************************
*                                                                            *
*     VITO -- Verilog Implicit To One-hot translator                         *
*     $Id: legal.txt,v 1.7 1999/06/03 15:24:08 mwinkel Apr_19_2004_Version_1.5 $
*                                                                            *
*     This software translates implicit style Verilog into a form suitable   *
*     for most commercial synthesis tools.  The translation is achieved      *
*     by constructing a one-hot machine for the ASM described in the         *
*     implicit style Verilog.                                                *
*                                                                            *
*     This work is being performed at the University of Wyoming and SUNY     *
*     College at Brockport.  For more information:                           *
*	   www.verilog.vito.com                                              *
*          http://strawberry.uwyo.edu/~vito                                  *
*          http://cs.brockport.edu/~jshuler/vito                             *
*                                                                            *
*     Authors:                                                               *
*          James D. Shuler  jshuler@cs.brockport.edu                         *
*          Mark G. Arnold   marnold@uwyo.edu                                 *
*                                                                            *
*                         NOTICE AND DISCLAIMER                              *
*                                                                            *
* No use of this software is authorized except under this NOTICE AND         *
* DISCLAIMER.  Downloading, use, copying, modification, and/or               *
* distribution of this software in its original or modified form implies     *
* your acceptance of this NOTICE and DISCLAIMER.                             *
*                                                                            *
* Permission to use, copy, modify and distribute this software and its       *
* documentation, for any purpose and without fee is hereby granted,          *
* provided that this permission notice appears prominently in supporting     *
* documentation and must be viewed prior to use, copying, modification or    *
* distribution.  You are responsible for any modifications to the software   *
* which you make and notice that the software has been modified must be      *
* appended to this notice prior to further use, copying, modification or     *
* distribution.                                                              *
*                                                                            *
* This software is provided AS IS without warranty of any kind,              *
* including without limitation the warranties that the software is           *
* non-infringing, merchantable, or fit for a particular purpose, including   *
* high risk activities.  The entire risk as to the quality and performance   *
* of the software is born by you.  Should the software prove defective in    *
* any respect, you and not the developers, nor any parties associated with   *
* the developers, assume the entire cost of  any service and repair.         *
*                                                                            *
* This software may be subject to the Export Control Laws of the United      *
* States of America.  It is your responsibility to determine the             *
* applicable laws and regulations and comply with them.                      *
*                                                                            *
*****************************************************************************/
#include "vito.h"

#define s_name			rc_string[WIREPRE]
#define edgeevent_name		rc_string[ATEVENT]
#define newreg_name		rc_string[NEWWIREPRE]
#define join_name		rc_string[JOINPRE]

/************************************************************************
$Id: vitofork.c,v 1.5 1999/06/03 15:54:53 mwinkel Apr_19_2004_Version_1.5 $

			f o r k _ f i l t e r

	Modifies every occurance of the fork statement in the v.out temporary
file into something understood by come_from.

************************************************************************/
void fork_filter(FILE **arg_out_file,FILE **arg_arch_file,int no_modules)
{
  int n;            /*statement number*/
  char s[80];       /*original statement*/
  int joinnum;      /*seq join_ id #*/
  int joinall;      /*join_ id of result of &*/
  int c;            /*count of begin, end, fork, join*/
  int i;
  int cur_module;
  FILE *temp_out_file,*new_out_file,*new_arch_file;
  FILE *out_file,*arch_file;

  out_file = *arg_out_file;
  arch_file = *arg_arch_file;
  new_out_file = tmpfile();
  new_arch_file = tmpfile();
  temp_out_file = tmpfile();
  if (new_out_file==NULL || new_arch_file==NULL || temp_out_file==NULL) {
    error("Cannot create a temporary file.\n");
    exit(EXIT_FAILURE);
  }

  for (cur_module=0;cur_module<no_modules;cur_module++){

  copy_file_module(new_arch_file,arch_file);

  joinnum = 0;
  n = 0;
  while (!feof(out_file))
    {
      fgets(s,80,out_file);
      n = n + 1;
      if (feof(out_file) || strcmp(s,MODULE_SEPARATOR)==0)
	 break;
	  if (strstr(s,"fork")==s)
	    {
	      fprintf(new_out_file,"//fork\n");
	      joinnum = joinnum + 1;
	      joinall = joinnum;
	      c = 1;
	      while ((c != 0) && (!feof(out_file)))
		{
		  fgets(s,80,out_file);
		  if (!feof(out_file))
		    {
		      if ((c == 1) && (strstr(s,"begin")==s))
			{
			  joinnum  = joinnum + 1;
			  fprintf(new_arch_file, "reg %s%d;\n",join_name,joinnum);
			  fprintf(new_arch_file, "wire %s%s%d;\n",newreg_name,join_name,joinnum);
			  fprintf(temp_out_file, "always\n");
			  fprintf(temp_out_file, "begin\n");
			  fprintf(temp_out_file, "%s%d<=%s0;\n",join_name,joinnum,edgeevent_name );
			  fprintf(temp_out_file, "%s #1 \n",edgeevent_name );
			  fprintf(temp_out_file, "if( %s%d )\n",s_name,n );
			  fprintf(temp_out_file, "begin\n");
			}
		      if ((strstr(s,"fork")==s) || (strstr(s,"begin")==s))
			c = c + 1;
		      if (c >= 2)
			  fprintf(temp_out_file, "%s",s);
		      if ((strstr(s,"join")==s) || (strstr(s,"end")==s))
			c = c - 1;
		      if ((c == 1) && (strstr(s,"end")==s))
			{
			  fprintf(temp_out_file, "while(~%s%d)\n",join_name,joinall );
			  fprintf(temp_out_file, "begin\n");
			  fprintf(temp_out_file, "%s%d<=%s1;\n",join_name, joinnum ,edgeevent_name);
			  fprintf(temp_out_file, "%s #1;\n",edgeevent_name);
			  fprintf(temp_out_file, "end\n");
			  fprintf(temp_out_file, "end\n");
			  fprintf(temp_out_file, "end\n");
			}
		    }
		}
	      if (feof(out_file))
		error("fork: early eof error\n");
      fgets(s,80,out_file);
      if (!feof(out_file))
      {
        if ((strstr(s,edgeevent_name)!=s))
         {
           error("fork: must have %s after join\n",edgeevent_name);
         }
      }
      else
		error("fork: early eof error\n");
      
	      fprintf(new_arch_file, "wire %s%d = %s%d",join_name,joinall,join_name,joinall+1);
	      for (i = joinall+2; i<= joinnum; i++)
		{
		  fprintf(new_arch_file,"&%s%d",join_name,i);
		}
	      fprintf(new_arch_file,";\n");
	      fprintf(new_out_file,"while(~%s%d)\n",join_name,joinall);
	      fprintf(new_out_file,"begin\n");
	      fprintf(new_out_file,"%s #1;\n",edgeevent_name);
	      fprintf(new_out_file,"end\n");
	      n = n + 4;
	    }
	  else
	    {
	      fprintf(new_out_file,"%s",s);
	    }
    }

  /* append the new always blocks to the out file */
  rewind(temp_out_file);
  copy_file(new_out_file,temp_out_file);
  fclose(temp_out_file);
  fprintf(new_out_file,"%s",MODULE_SEPARATOR);
  fprintf(new_arch_file,"%s",MODULE_SEPARATOR);

  } /* for each module */

  fclose(out_file);
  fclose(arch_file);
  *arg_out_file = new_out_file;
  *arg_arch_file = new_arch_file;
}
