/*****************************************************************************
*                                                                            *
*     VITO -- Verilog Implicit To One-hot translator                         *
*     $Id: legal.txt,v 1.7 1999/06/03 15:24:08 mwinkel Apr_19_2004_Version_1.5 $
*                                                                            *
*     This software translates implicit style Verilog into a form suitable   *
*     for most commercial synthesis tools.  The translation is achieved      *
*     by constructing a one-hot machine for the ASM described in the         *
*     implicit style Verilog.                                                *
*                                                                            *
*     This work is being performed at the University of Wyoming and SUNY     *
*     College at Brockport.  For more information:                           *
*	   www.verilog.vito.com                                              *
*          http://strawberry.uwyo.edu/~vito                                  *
*          http://cs.brockport.edu/~jshuler/vito                             *
*                                                                            *
*     Authors:                                                               *
*          James D. Shuler  jshuler@cs.brockport.edu                         *
*          Mark G. Arnold   marnold@uwyo.edu                                 *
*                                                                            *
*                         NOTICE AND DISCLAIMER                              *
*                                                                            *
* No use of this software is authorized except under this NOTICE AND         *
* DISCLAIMER.  Downloading, use, copying, modification, and/or               *
* distribution of this software in its original or modified form implies     *
* your acceptance of this NOTICE and DISCLAIMER.                             *
*                                                                            *
* Permission to use, copy, modify and distribute this software and its       *
* documentation, for any purpose and without fee is hereby granted,          *
* provided that this permission notice appears prominently in supporting     *
* documentation and must be viewed prior to use, copying, modification or    *
* distribution.  You are responsible for any modifications to the software   *
* which you make and notice that the software has been modified must be      *
* appended to this notice prior to further use, copying, modification or     *
* distribution.                                                              *
*                                                                            *
* This software is provided AS IS without warranty of any kind,              *
* including without limitation the warranties that the software is           *
* non-infringing, merchantable, or fit for a particular purpose, including   *
* high risk activities.  The entire risk as to the quality and performance   *
* of the software is born by you.  Should the software prove defective in    *
* any respect, you and not the developers, nor any parties associated with   *
* the developers, assume the entire cost of  any service and repair.         *
*                                                                            *
* This software may be subject to the Export Control Laws of the United      *
* States of America.  It is your responsibility to determine the             *
* applicable laws and regulations and comply with them.                      *
*                                                                            *
*****************************************************************************/

DIGIT	[0-9]
LETTER	[a-zA-Z]
UNUM	({DIGIT}({DIGIT}|_)*)
WS	(" "|\t|"")
%pointer

%{
/*****************************************************************************
$Id: vitol.l,v 1.5 1999/06/03 15:55:38 mwinkel Apr_19_2004_Version_1.5 $

	Lexical scanner for VITO using lex.  Keywords not supported by
   VITO are not recognized by this scanner.

*****************************************************************************/
#include "vito.h"
#include "vitop.h"

%}

%Start	COMMENT

%e 2000
%p 7000
%a 4000
%o 6000


%%
\\[^ \t\n]*({WS}|\n)		{ /* Escaped Identifier */
			  	   /*tprint(yytext);*/
				   if (yytext[yyleng-1]=='\n')
				      line_no++;
			   	   yylval.identifier = Insert_ID(yytext,yyleng);
			   	   found_token(); return(XIDENTIFIER); }
"/*"				{ /* Beginning of long comment */
				   BEGIN COMMENT;
				   comment(yytext);
				   continue; }
<COMMENT>[^*\n]*\n?		{ /* Non '*' text of long comment */
				   if (yytext[yyleng-1]=='\n')
				      line_no++;
				   comment(yytext);
				   continue; }
<COMMENT>"*"/[^/]\n?		{ /* '*' text of long comment */
				   if (yytext[yyleng-1]=='\n')
				      line_no++;
				   comment(yytext);
				   continue; }
<COMMENT>"*/"\n?		{ /* Termination of long comment */
				   if (yytext[yyleng-1]=='\n')
				      line_no++;
				   BEGIN 0;
				   comment(yytext);
				   continue; }
<COMMENT><<EOF>>		{  error("Unterminated comment.\n");
				   exit(EXIT_FAILURE);
				}
"//".*\n			{ /* short comment */
				   if (yytext[yyleng-1]=='\n')
				      line_no++;
				   comment(yytext);
				   continue; }
^{WS}*"`define".*\n		{ /* Macro definition */
				   if (yytext[yyleng-1]=='\n')
				      line_no++;
				   /*
				   put_stmt(yytext);
				   */
				   comment(yytext);
				 }

{WS}+				{ /* Skip white space */
				   comment(yytext);
				   continue; }
\n				{ /* Skip white space */
				   /*
				   if (yytext[yyleng-1]=='\n')
				   */
				      line_no++;
				   comment(yytext);
				   continue; }

always				{ found_token(); return(XALWAYS); }
and				{ found_token(); return(XAND); }
assign				{ found_token(); return(XASSIGN); }
begin				{ found_token(); return(XBEGIN); }
buf				{ found_token(); return(XBUF); }
bufif0				{ found_token(); return(XBUFIF0); }
bufif1				{ found_token(); return(XBUFIF1); }
case				{ found_token(); return(XCASE); }
casex				{ found_token(); return(XCASEX); }
casez				{ found_token(); return(XCASEZ); }
cmos				{ found_token(); return(XCMOS); }
deassign			{ found_token(); return(XDEASSIGN); }
default				{ found_token(); return(XDEFAULT); }
defparam			{ found_token(); return(XDEFPARAM); }
disable				{ found_token(); return(XDISABLE); }
else				{ found_token(); return(XELSE); }
end				{ found_token(); return(XEND); }
endcase				{ found_token(); return(XENDCASE); }
endfunction			{ found_token(); return(XENDFUNCTION); }
endmodule			{ found_token(); return(XENDMODULE); }
endtask				{ found_token(); return(XENDTASK); }
event				{ found_token(); return(XEVENT); }
for				{ found_token(); return(XFOR); }
force				{ found_token(); return(XFORCE); }
forever				{ found_token(); return(XFOREVER); }
fork				{ found_token(); return(XFORK); }
function			{ found_token(); return(XFUNCTION); }
highz0				{ found_token(); return(XHIGHZ0); }
highz1				{ found_token(); return(XHIGHZ1); }
if				{ found_token(); return(XIF); }
initial				{ found_token(); return(XINITIAL); }
inout				{ found_token(); return(XINOUT); }
input				{ found_token(); return(XINPUT); }
integer				{ found_token(); return(XINTEGER); }
join				{ found_token(); return(XJOIN); }
large				{ found_token(); return(XLARGE); }
medium				{ found_token(); return(XMEDIUM); }
module				{ found_token(); return(XMODULE); }
nand				{ found_token(); return(XNAND); }
negedge				{ found_token(); return(XNEGEDGE); }
nmos				{ found_token(); return(XNMOS); }
nor				{ found_token(); return(XNOR); }
not				{ found_token(); return(XNOT); }
notif0				{ found_token(); return(XNOTIF0); }
notif1				{ found_token(); return(XNOTIF1); }
or				{ found_token(); return(XOR); }
output				{ found_token(); return(XOUTPUT); }
parameter			{ found_token(); return(XPARAMETER); }
pmos				{ found_token(); return(XPMOS); }
posedge				{ found_token(); return(XPOSEDGE); }
pull0				{ found_token(); return(XPULL0); }
pull1				{ found_token(); return(XPULL1); }
pulldown			{ found_token(); return(XPULLDOWN); }
pullup				{ found_token(); return(XPULLUP); }
rcmos				{ found_token(); return(XRCMOS); }
real				{ found_token(); return(XREAL); }
realtime			{ found_token(); return(XREALTIME); }
reg				{ found_token(); return(XREG); }
release				{ found_token(); return(XRELEASE); }
repeat				{ found_token(); return(XREPEAT); }
rnmos				{ found_token(); return(XRNMOS); }
rpmos				{ found_token(); return(XRPMOS); }
rtran				{ found_token(); return(XRTRAN); }
rtranif0			{ found_token(); return(XRTRANIF0); }
rtranif1			{ found_token(); return(XRTRANIF1); }
scalared			{ found_token(); return(XSCALARED); }
small				{ found_token(); return(XSMALL); }
strong0				{ found_token(); return(XSTRONG0); }
strong1				{ found_token(); return(XSTRONG1); }
supply0				{ found_token(); return(XSUPPLY0); }
supply1				{ found_token(); return(XSUPPLY1); }
task				{ found_token(); return(XTASK); }
time				{ found_token(); return(XTIME); }
tran				{ found_token(); return(XTRAN); }
tranif0				{ found_token(); return(XTRANIF0); }
tranif1				{ found_token(); return(XTRANIF1); }
tri				{ found_token(); return(XTRI); }
tri0				{ found_token(); return(XTRI0); }
tri1				{ found_token(); return(XTRI1); }
triand				{ found_token(); return(XTRIAND); }
trior				{ found_token(); return(XTRIOR); }
trireg				{ found_token(); return(XTRIREG); }
vectored			{ found_token(); return(XVECTORED); }
wait				{ found_token(); return(XWAIT); }
wand				{ found_token(); return(XWAND); }
wor				{ found_token(); return(XWOR); }
weak0				{ found_token(); return(XWEAK0); }
weak1				{ found_token(); return(XWEAK1); }
while				{ found_token(); return(XWHILE); }
wire				{ found_token(); return(XWIRE); }
xor				{ found_token(); return(XXOR); }
xnor				{ found_token(); return(XXNOR); }
"{"				{ found_token(); return(XLBRACKET); }
"}"				{ found_token(); return(XRBRACKET); }
","				{ found_token(); return(XCOMMA); }
"+"				{ found_token(); return(XPLUS); }
"-"				{ found_token(); return(XMINUS); }
"*"				{ found_token(); return(XMULT); }
"/"				{ found_token(); return(XDIVIDE); }
"%"				{ found_token(); return(XMOD); }
">"				{ found_token(); return(XGREATER); }
">="				{ found_token(); return(XGREATEREQ); }
"<"				{ found_token(); return(XLESS); }
"<="				{ found_token(); return(XLESSEQ); }
"->"				{ found_token(); return(XEVENTTRIGGER); }
"!"				{ found_token(); return(XLNOT); }
"&&"				{ found_token(); return(XLAND); }
"||"				{ found_token(); return(XLOR); }
"=="				{ found_token(); return(XLEQUAL); }
"!="				{ found_token(); return(XLNOTEQUAL); }
"==="				{ found_token(); return(XCEQUAL); }
"!=="				{ found_token(); return(XCNOTEQUAL); }
"~"				{ found_token(); return(XBNOT); }
"&"				{ found_token(); return(XBAND); }
"~&"				{ found_token(); return(XBNAND); }
"|"				{ found_token(); return(XBOR); }
"~|"				{ found_token(); return(XBNOR); }
"^"				{ found_token(); return(XBXOR); }
"^~"|"~^"			{ found_token(); return(XBXNOR); }
"<<"				{ found_token(); return(XLSHIFT); }
">>"				{ found_token(); return(XRSHIFT); }
"?"				{ found_token(); return(XQUESTION); }
":"				{ found_token(); return(XCOLON); }
";"				{ found_token(); return(XSEMI); }
"("				{ found_token(); return(XLPAREN); }
")"				{ found_token(); return(XRPAREN); }
"["				{ found_token(); return(XLSQUARE); }
"]"				{ found_token(); return(XRSQUARE); }
"#"				{ found_token(); return(XPOUND); }
"."				{ found_token(); return(XPERIOD); }
"$"				{ found_token(); return(XDOLLAR); }
"@"				{ found_token(); return(XAT); }
"="				{ found_token(); return(XEQUAL); }

[a-zA-Z_][a-zA-Z_$0-9]*		{ /* simple identifier */
			   	   yylval.identifier = Insert_ID(yytext,yyleng);
			   	   found_token(); return(XIDENTIFIER); }
\"([^"\n]|\\\")*\"		{ /* String */
			   	   yylval.identifier = Insert_ID(yytext,yyleng);
			   	   found_token(); return(XSTRING); }
({UNUM})|({UNUM}?'[dD]{UNUM})	{ /* Decimal number */
			           yylval.identifier = Insert_ID(yytext,yyleng);
				   found_token(); return(XNUMBER); }
{UNUM}?'[bB][0-1xXzZ][0-1xXzZ_]*	{ /* Binary number */
			           yylval.identifier = Insert_ID(yytext,yyleng);
				   found_token(); return(XNUMBER); }
{UNUM}?'[oO][0-7xXzZ][0-7xXzZ_]*	{ /* Octal number */
			           yylval.identifier = Insert_ID(yytext,yyleng);
				   found_token(); return(XNUMBER); }
{UNUM}?'[hH][0-9a-fA-FxXzZ][0-9a-fA-FxXzZ_]*	{ /* Hex number */
			           yylval.identifier = Insert_ID(yytext,yyleng);
				   found_token(); return(XNUMBER); }
{UNUM}("."{UNUM}|("."{UNUM})?e[+-]?{UNUM})	{ /* Real number */
			           yylval.identifier = Insert_ID(yytext,yyleng);
				   found_token(); return(XNUMBER); }

.				{ /* Any other Character */
				   yyerror(yytext);
				}
%%
