/*****************************************************************************
*                                                                            *
*     VITO -- Verilog Implicit To One-hot translator                         *
*     $Id: legal.txt,v 1.7 1999/06/03 15:24:08 mwinkel Apr_19_2004_Version_1.5 $
*                                                                            *
*     This software translates implicit style Verilog into a form suitable   *
*     for most commercial synthesis tools.  The translation is achieved      *
*     by constructing a one-hot machine for the ASM described in the         *
*     implicit style Verilog.                                                *
*                                                                            *
*     This work is being performed at the University of Wyoming and SUNY     *
*     College at Brockport.  For more information:                           *
*	   www.verilog.vito.com                                              *
*          http://strawberry.uwyo.edu/~vito                                  *
*          http://cs.brockport.edu/~jshuler/vito                             *
*                                                                            *
*     Authors:                                                               *
*          James D. Shuler  jshuler@cs.brockport.edu                         *
*          Mark G. Arnold   marnold@uwyo.edu                                 *
*                                                                            *
*                         NOTICE AND DISCLAIMER                              *
*                                                                            *
* No use of this software is authorized except under this NOTICE AND         *
* DISCLAIMER.  Downloading, use, copying, modification, and/or               *
* distribution of this software in its original or modified form implies     *
* your acceptance of this NOTICE and DISCLAIMER.                             *
*                                                                            *
* Permission to use, copy, modify and distribute this software and its       *
* documentation, for any purpose and without fee is hereby granted,          *
* provided that this permission notice appears prominently in supporting     *
* documentation and must be viewed prior to use, copying, modification or    *
* distribution.  You are responsible for any modifications to the software   *
* which you make and notice that the software has been modified must be      *
* appended to this notice prior to further use, copying, modification or     *
* distribution.                                                              *
*                                                                            *
* This software is provided AS IS without warranty of any kind,              *
* including without limitation the warranties that the software is           *
* non-infringing, merchantable, or fit for a particular purpose, including   *
* high risk activities.  The entire risk as to the quality and performance   *
* of the software is born by you.  Should the software prove defective in    *
* any respect, you and not the developers, nor any parties associated with   *
* the developers, assume the entire cost of  any service and repair.         *
*                                                                            *
* This software may be subject to the Export Control Laws of the United      *
* States of America.  It is your responsibility to determine the             *
* applicable laws and regulations and comply with them.                      *
*                                                                            *
*****************************************************************************/

DIGIT	[0-9]
LETTER	[a-zA-Z]
UNUM	({DIGIT}({DIGIT}|_)*)
WS	(" "|\t|"")
%pointer

%{
/*****************************************************************************
$Id: vitopp.l,v 1.5 1999/06/03 15:57:00 mwinkel Apr_19_2004_Version_1.5 $

	Implements a simple preprocessor for Verilog.  It concatentates
lines terminated by '\' and expands macros defined by '`define'.  The original
'`define' is left in place.

*****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "vito.h"

#define TRUE	1
#define FALSE	0
/*  Tokens  */
#define TIDENTIFIER	1
#define TSTRING		2
#define TDEFINE		3
#define TNEWLINE	4
#define TQUOTE		6
#define TWHITESPACE	7
#define TCHAR		8
#define TIFDEF		9
#define TELSE		10
#define TENDIF		11
#define TINCLUDE	12
#define TEOF		100

int lex_ID;
int last_macro_pp;
int token;
struct {
   YY_BUFFER_STATE buffer;
   FILE * file;
}buffer_stack[MAX_MACRO_DEPTH];
int buffer_sp;
int macro_depth;
#if FALSE
struct {
   char *ID;		/* the identifier string */
   char *macro;		/* the macro string (NULL if none) */
   int perm;		/* TRUE if macro should not be redefinable */
   #endif
macro_table_type macro_table_pp[MAXID];
int destroy_comments;	/* TRUE if comments are to be destroyed (not output) */

void put_out_pp(char *string);
void process_token(int token);
%}

%Start	COMMENT

%e 2000
%p 7000
%a 3000
%o 5000


%%
"/*"			{ /* Beginning of long comment */
			   BEGIN COMMENT;
			   if (!destroy_comments)
			      put_out_pp(yytext);
			   continue; }
<COMMENT>[^*\n]*\n?	{ /* Non '*' text of long comment */
			   if(yytext[yyleng-1]=='\n')
			      line_no++;
			   if (!destroy_comments)
			      put_out_pp(yytext);
			   continue; }
<COMMENT>"*"/[^/]\n?	{ /* '*' text of long comment */
			   if(yytext[yyleng-1]=='\n')
			      line_no++;
			   if (!destroy_comments)
			      put_out_pp(yytext);
			   continue; }
<COMMENT>"*/"		{ /* Termination of long comment */
			   /* \n?
			   if(yytext[yyleng-1]=='\n')
			      line_no++;
			      */
			   BEGIN 0;
			   if (!destroy_comments)
			      put_out_pp(yytext);
			   continue; }
"//"([^\\\n]|\\.|\\\n)*			{ /* short comment */
			   /* /n
			   if(yytext[yyleng-1]=='\n')
			      line_no++;
			      */
			   if (!destroy_comments)
			      put_out_pp(yytext);
			   continue; }

({WS})*			{ return TWHITESPACE; }

[a-zA-Z_][a-zA-Z_$0-9]*	{ /* simple identifier */
			   /*lex_ID = Insert_ID_pp(yytext,yyleng);*/
			   return TIDENTIFIER; }

\"([^"\n]|\\\")*\"	{ /* String */
			   /*lex_ID = Insert_ID_pp(yytext,yyleng);*/
			   return TSTRING; }

`define			{ return TDEFINE; }
`ifdef			{ return TIFDEF; }
`else			{ return TELSE; }
`endif			{ return TENDIF; }
`include		{ return TINCLUDE; }

`			{ return TQUOTE; }

\n			{ line_no++;
			  return TNEWLINE; }

\\\n			{ line_no++;
			  yytext[0]=' ';
			  yytext[1]= '\0';
			  return TWHITESPACE; }
\032			{ /* ignore DOS EOF */ }
\015			{ /* ignore DOS EOLN */ }
.			{ return TCHAR;}

<<EOF>>			{  
			      yyterminate();
			   if (macro_depth > 0)
      			      return TEOF;
      			   else{
			      yyterminate();
      			      }}
%%
void push_buffer(YY_BUFFER_STATE buffer,FILE *file_ptr)
{
   if (buffer_sp >= MAX_MACRO_DEPTH) {
      error("Too many recursive macros, includes, or conditional directives.  Increase MAX_MACRO_DEPTH.\n");
      exit(EXIT_FAILURE);
   }
   buffer_stack[buffer_sp].buffer = buffer;
   buffer_stack[buffer_sp].file = file_ptr;
   buffer_sp++;
}
YY_BUFFER_STATE pull_buffer(FILE **file_ptr)
{
   if (buffer_sp<=0) {
      error("Underflow on the buffer_stack!\n");
      exit(EXIT_FAILURE);
   }
   buffer_sp--;
   *file_ptr = buffer_stack[buffer_sp].file;
   return buffer_stack[buffer_sp].buffer;
}
int empty_buffer(void)
{
   return buffer_sp==0;
}
/*
void increase_macro_depth(void)
{
   macro_depth++;
   if (macro_depth>MAX_MACRO_DEPTH) {
      error("Too many recursive macros, includes, or conditional directives.  Increase MAX_MACRO_DEPTH.\n");
      exit(EXIT_FAILURE);
   }
}
void decrease_macro_depth(void)
{
   if (macro_depth<=0)
      error("macro_depth less than zero!\n");
   macro_depth--;
}
*/
/*****************************************************************************

			y y w r a p

	 Lex/YACC function to obtain next file for input.

*****************************************************************************/
int yywrap(void)
{
   static cur_file=0;
   FILE *file_ptr;

   /*
   if (macro_depth > 0)
      return TRUE;
      */
   if (!empty_buffer()){
      yy_delete_buffer(YY_CURRENT_BUFFER);
      yy_switch_to_buffer(pull_buffer(&file_ptr));
      if (file_ptr!=NULL)
	 if (fclose(file_ptr)!=0)
	    error("Could not close an `include file properly.\n");
      return FALSE;
   }
   fclose(yyin);
   cur_file++;
   if (cur_file < last_source_file) {
      yyin = fopen(source_files[cur_file],"r");
      if (yyin==NULL) {
	 error("Cannot open file '%s'.\n",source_files[cur_file]);
	 exit(EXIT_FAILURE);
      }
      return FALSE;
   }
   else
      return TRUE;
}
/*****************************************************************************

			e r r o r

	 Prints an error message indicating the name of this tool and the line
   number.

Input:	string,... = the specific message to print

*****************************************************************************/
/*
void error(char *string, ...)
{
   va_list ap;
   va_start(ap,string);
   fprintf(stderr,"%s Error %d:  ",NAME_OF_TOOL,line_no);
   vfprintf(stderr,string,ap);
   no_errors++;
}
*/
/*****************************************************************************

			i n f o

	 Prints an information message indicating the name of this tool.

Input:	string,... = the specific message to print

*****************************************************************************/
/*
void info(char *string, ...)
{
   va_list ap;
   va_start(ap,string);
   fprintf(stderr,"%s:  ",NAME_OF_TOOL);
   vfprintf(stderr,string,ap);
}
*/
/*****************************************************************************

			i n i t i a l i z e _ p p

	 Initializes global variables.

*****************************************************************************/
void initialize_pp(void)
{
   int cur_id;

   destroy_comments=FALSE;
   line_no= 1;
   buffer_sp=0;
   /*
   for (cur_id=0;cur_id<MAXID;cur_id++) {
      macro_table_pp[cur_id].ID = NULL;
      macro_table_pp[cur_id].macro = NULL;
   }
   last_macro_pp= 0; Done in vito.c now
   */
   /* open first source file if given */
   if (last_source_file >0 ) {
      fclose(yyin);
      yyin = fopen(source_files[0],"r");
      if (yyin==NULL) {
	 error("Cannot open file '%s'.\n",source_files[0]);
	 exit(EXIT_FAILURE);
      }
   }
}
/*****************************************************************************

			G e t _ I D _ p p

      Obtains the identifier stored at location position in the macro_table_pp.

Input:	location = index to macro_table_pp
Output:	Get_ID_pp = indentifier string at location

*****************************************************************************/
char *Get_ID_pp(int location)
{
   return macro_table_pp[location].ID;
}
/*****************************************************************************


			F i n d _ I D _ p p

	 Given an identifier, it finds its entry in the macro_table_pp.  If not
found it returns a negative number.

Input:	ID = identifier to search for
Output:	Find_ID_pp = index in macro_table_pp or -1

*****************************************************************************/
int Find_ID_pp(char *ID)
{
   int cur_loc;
   for(cur_loc=0;cur_loc<last_macro_pp;cur_loc++)
      if(strcmp(ID,macro_table_pp[cur_loc].ID)==0)
	 return cur_loc;
   return -1;
}
/*****************************************************************************

			I n s e r t _ I D _ p p

	 Inserts the given identifier into the macro_table_pp.  If it is already in
the table, that location is returned.  Currently implemented using a linear
search technique but ideally would use something more sophisticated such
as a hash table.

Input:	ID = the identifier to insert (must contain the \0)
	length = the number of characters in the identifier (without \0)
Output:	Insert_ID_pp = index of location in macro_table_pp

*****************************************************************************/
int Insert_ID_pp(char *ID,int length)
{
   int loc;
   if (last_macro_pp >= MAXID-1) {
      error("Exceeded internal ID table.  Increase MAXID.\n");
      exit(EXIT_FAILURE);
   }
   loc=Find_ID_pp(ID);
   if (loc>=0)
      return loc;
   macro_table_pp[last_macro_pp].ID = (char *)malloc(length+1);
   if (macro_table_pp[last_macro_pp].ID==NULL) {
      error("Malloc returned NULL.  Run out of memory.\n");
      exit(EXIT_FAILURE);
   }
   macro_table_pp[last_macro_pp].macro = NULL;
   macro_table_pp[last_macro_pp].perm = FALSE;
   strcpy(macro_table_pp[last_macro_pp].ID,ID);
   last_macro_pp++;
   return (last_macro_pp-1);
}
/*****************************************************************************

			p u t _ o u t _ p p

      Outputs the given string to be passed on to the next stage.
Currently just goes to standard output.

Input:	string = string to print

*****************************************************************************/
void put_out_pp(char *string)
{
   fprintf(yyout,"%s",string);
}
/*****************************************************************************

			e x p a n d _ m a c r o

      Expands a macro found in normal text.  If it has not been defined an
error message is printed.

*****************************************************************************/
void expand_macro(void)
{
   int cur_ID;
   int token;
   YY_BUFFER_STATE macro_state;

   token=yylex();
   if (token==TIDENTIFIER) {
      cur_ID = Insert_ID_pp(yytext,yyleng);
      if (macro_table_pp[cur_ID].macro==NULL) {
	 error("Macro '`%s' undefined.\n",yytext);
	 put_out_pp("`");
	 put_out_pp(yytext);
      }
      else {
	 /*macro_state = YY_CURRENT_BUFFER;*/
	 push_buffer(YY_CURRENT_BUFFER,NULL);
	 /*increase_macro_depth();*/
	 yy_switch_to_buffer(yy_scan_string(macro_table_pp[cur_ID].macro));
	 /*
	 token = yylex();
	 while (token > 0 && token!=TEOF) {
	    process_token(token);
	    token=yylex();
	 }
	 yy_switch_to_buffer(macro_state);
	 decrease_macro_depth();
	 */
      }
   }
   else {
      put_out_pp("`");
      put_out_pp(yytext);
   }
}
/*****************************************************************************

			m y s t r c a t

      Concatenates the second string onto the first with error checking.
If there is insufficient storage, then an error is printed and processing
resumes with a single error reported for the line.

Input:	s1 = the string to concatenate to
	s2 = the new string to concatenate onto s1

*****************************************************************************/
void mystrcat(char *s1,char *s2)
{
   int l1;
   static last_error_line= -1;

   l1=strlen(s1);
   if (l1>= MAXMACRO) {
      if (last_error_line==line_no)
	 return;
      last_error_line= line_no;
      error("Exceeded internal size for a macro.  Increase MAXMACRO.\n");
      return;
      /*exit(EXIT_FAILURE);*/
   }
   strncat(s1,s2,MAXMACRO-l1-1);
   if (l1 + strlen(s2) >= MAXMACRO) {
      if (last_error_line==line_no)
	 return;
      last_error_line= line_no;
      error("Exceeded internal size for a macro.  Increase MAXMACRO.\n");
      /*exit(EXIT_FAILURE);*/
   }
}
void define_macro(void)
{
   int cur_ID;
   char macro[MAXMACRO];

   put_out_pp(yytext);
   token=yylex();
   if (token==TWHITESPACE) {
      put_out_pp(yytext);
      token=yylex();
      if (token==TIDENTIFIER) {
         cur_ID = Insert_ID_pp(yytext,yyleng);
	 if (!macro_table_pp[cur_ID].perm){
            put_out_pp(yytext);
            /* remove any prior definition */
            if (macro_table_pp[cur_ID].macro != NULL)
               free(macro_table_pp[cur_ID].macro);
	 }
	 else {
	    warning("Ignoring redefinition of macro '%s' (command line defined).\n",macro_table_pp[cur_ID].ID);
	    put_out_pp(macro_table_pp[cur_ID].ID);
	    put_out_pp(" ");
	    put_out_pp(macro_table_pp[cur_ID].macro);
	 }
         /* skip whitespace */
         token=yylex();
         if (token==TWHITESPACE) {
	    if (!macro_table_pp[cur_ID].perm)
               put_out_pp(yytext);
            token=yylex();
         }
         /* read in new definition */
         macro[0]='\0';
         while (token!=TNEWLINE) {
	    if (!macro_table_pp[cur_ID].perm) {
               mystrcat(macro,yytext);
               put_out_pp(yytext);
	    }
            token=yylex();
         } /* while */
         put_out_pp(yytext);
	 if (!macro_table_pp[cur_ID].perm) { 
            macro_table_pp[cur_ID].macro = (char *)malloc(strlen(macro)+1);
            if (macro_table_pp[cur_ID].macro==NULL) {
               error("Malloc returned NULL.  Ran out of memory.\n");
               exit(EXIT_FAILURE);
            }
            strcpy(macro_table_pp[cur_ID].macro,macro);
	 }
      }
      else {
         /* non identifier after `define */
         error("No identifier following `define.\n");
      }
   }
   else {
      /* non whitespace after `define */
      error("Improper `define.\n");
   }
}
void conditional(void)
{
   int token;
   int cur_ID;

   destroy_comments=TRUE;
   token=yylex();
   if (token==TWHITESPACE) {
      do
         token=yylex();
      while (token==TWHITESPACE);
      if (token==TIDENTIFIER) {
	 cur_ID = Find_ID_pp(yytext);
         if (cur_ID<0 || macro_table_pp[cur_ID].macro == NULL) {
            /* macro not defined...skip to `else */
            token=yylex();
            while (token != TELSE && token!=TENDIF && token>0) {
               token=yylex();
            }
            if (token <= 0) {
               error("End of file reached without `ELSE or `ENDIF completing an `IFDEF\n");
               exit(EXIT_FAILURE);
            }
            if (token==TELSE) {
	       destroy_comments=FALSE;
               token==yylex();
               while (token!=TENDIF && token>0) {
                  process_token(token);
	          destroy_comments=FALSE;
                  token=yylex();
               }
               if (token <= 0) {
                  error("End of file reached without `ENDIF completing an `IFDEF\n");
                  exit(EXIT_FAILURE);
               }
            }
         }
         else {
            /* macro defined */
	    destroy_comments=FALSE;
            token=yylex();
            while (token>0 && token != TELSE && token!=TENDIF) {
               process_token(token);
	       destroy_comments=FALSE;
               token=yylex();
            }
            if (token <= 0) {
               error("End of file reached without `ELSE or `ENDIF completing an `IFDEF\n");
               exit(EXIT_FAILURE);
            }
            if (token==TELSE) {
	       destroy_comments=TRUE;
               token==yylex();
               while (token!=TENDIF && token>0) {
                  token=yylex();
               }
               if (token <= 0) {
                  error("End of file reached without `ENDIF completing an `IFDEF\n");
                  exit(EXIT_FAILURE);
               }
            }
         }
	 /*
         do
            token=yylex();
         while (token==TWHITESPACE);
	 */
      }
      else {
         error("No identifier following `ifdef.\n");
      }
   }
   else {
      error("Expecting whitespace after `ifdef.\n");
   }
   destroy_comments=FALSE;
}
void include(void)
{
   int cur_ID;
   FILE *include_file;
   int cur_lib,libn;
   char *lib;

   destroy_comments=TRUE;
   do
      token=yylex();
   while (token==TWHITESPACE);
   if (token==TSTRING) {
      if (yytext[0]!='\"' || yytext[yyleng-1]!='\"' || yyleng < 2) {
	 error("Bad string given for filename of `include.\n");
         destroy_comments=FALSE;
	 return;
      }
      yytext[yyleng-1] = '\0';
      cur_ID = Insert_ID_pp(yytext+1,yyleng-2);
      /* skip to end of line */
      do {
         token=yylex();
	 if (token!= TWHITESPACE && token!=TNEWLINE)
	    error("Extraneous characters after `include.\n");
      } while (token!=TNEWLINE);
      /* open named file as new input */
      include_file = fopen(Get_ID_pp(cur_ID),"r");
      if (include_file==NULL) {
	 /* try the library paths */
	 cur_lib=0;
         while (library_paths[cur_lib]!='\0') {
	    libn=strcspn(library_paths+cur_lib," ");
	    lib = (char*)malloc(strlen(Get_ID_pp(cur_ID))+libn+2);
	    if (lib==NULL) {
	       error("Malloc failed.  Ran out of memory.\n");
	       exit(EXIT_FAILURE);
	    }
	    memcpy(lib,library_paths+cur_lib,libn);
	    lib[libn]='\0';
#if OPERATING_SYSTEM==UNIX
	    if (lib[libn-1]!='/')
	       strcat(lib,"/");
#elif OPERATING_SYSTEM==DOS
	    if (lib[libn-1]!='\\')
	       strcat(lib,"\\");
#endif
	    strcat(lib,Get_ID_pp(cur_ID));
            include_file = fopen(lib,"r");
	    free(lib);
	    if (include_file!=NULL)
	       break;
	    cur_lib += libn;
	    if (library_paths[cur_lib]==' ')
	       cur_lib++;
	 }
         if (include_file==NULL) {
	    error("Unable to open the `include file '%s'.\n",Get_ID_pp(cur_ID));
            destroy_comments=FALSE;
	    return;
         }
      }
      push_buffer(YY_CURRENT_BUFFER,include_file);
      yy_switch_to_buffer(yy_create_buffer(include_file,YY_BUF_SIZE));
   }
   else {
      /* non string after `include */
      error("No filename following `include.\n");
   }
   destroy_comments=FALSE;
}
/************************************************************************

			P r o c e s s _ t o k e n

************************************************************************/
void process_token(int token)
{
   int cur_ID;
   char macro[MAXMACRO];

   switch (token) {
      case TDEFINE:
	 define_macro();
	 break;
      case TQUOTE:
         expand_macro();
         break;
      case TIFDEF:
	 conditional();
         break;
      case TINCLUDE:
	 include();
	 break;
      default:
         put_out_pp(yytext);
   }
}
/*****************************************************************************

			v i t o p p

      Performs a preprocessing step on the source code.  This exapands macros.
It also combines all source files into a single temporary file.

Input:	ofile = the output temporary file

*****************************************************************************/
void vitopp(FILE *ofile)
{
   int cur_ID;

   initialize_pp();
   yyout = ofile;
   token=yylex();
   while (token>0) {
      process_token(token);
      token=yylex();
   }
#if FALSE      
      if (token==TDEFINE) {
	 put_out_pp(yytext);
	 token=yylex();
	 if (token==TWHITESPACE) {
	    put_out_pp(yytext);
	    token=yylex();
	    if (token==TIDENTIFIER) {
	       put_out_pp(yytext);
	       cur_ID = lex_ID;
	       if (macro_table_pp[cur_ID].macro != NULL)
		  free(macro_table_pp[cur_ID].macro);
	       macro_table_pp[cur_ID].macro = (char *)malloc(MAXMACRO);
               if (macro_table_pp[cur_ID].macro==NULL) {
                  error("Malloc returned NULL.  Run out of memory.\n");
                  exit(EXIT_FAILURE);
               }
	       macro_table_pp[cur_ID].macro[0] = '\0';
	       token=yylex();
	       if (token==TWHITESPACE) {
	          put_out_pp(yytext);
	          token=yylex();
	       }
	       while (token!=TNEWLINE) {
		  if (token==TQUOTE) {
	             put_out_pp(yytext);
                     token=yylex();
                     if (token==TIDENTIFIER) {
                        if (macro_table_pp[lex_ID].macro==NULL) {
                            error("Macro '%s' undefined.\n",yytext);
		            mystrcat(macro_table_pp[cur_ID].macro,"`");
		            mystrcat(macro_table_pp[cur_ID].macro,yytext);
                         }
                         else if (lex_ID==cur_ID) {
			    error("Cannot define a macro in terms of itself.\n");
		            mystrcat(macro_table_pp[cur_ID].macro,"`");
		            mystrcat(macro_table_pp[cur_ID].macro,yytext);
			 }
			 else
		            mystrcat(macro_table_pp[cur_ID].macro,macro_table_pp[lex_ID].macro);
                      }
                      else
		         mystrcat(macro_table_pp[cur_ID].macro,"`");
		  }
		  else
		     mystrcat(macro_table_pp[cur_ID].macro,yytext);
	          put_out_pp(yytext);
		  token=yylex();
	       }
	    }
	 }
      }
      /*
      if (token==TQUOTE)
	 expand_macro();
      else
         put_out_pp(yytext);
	 */
      switch (token) {
	 case TQUOTE:
	    expand_macro();
	    break;
	 case TIFDEF:
            token=yylex();
            if (token==TWHITESPACE) {
               token=yylex();
               if (token==TIDENTIFIER) {
                  cur_ID = lex_ID;
                  if (macro_table_pp[cur_ID].macro == NULL) {
		     /* macro not defined...skip to `else */
		     token=yylex();
		     while (token != TELSE && token!=TENDIF) {
                        put_out_pp(yytext);
			token=yylex();
		     }
		     if (token <= 0) {
			error("End of file reached without `ELSE or `ENDIF completing an `IFDEF\n");
			exit(EXIT_FAILURE);
		     }
		     if (token==TELSE) {
			token==yylex();
			while (token!=TENDIF) {
			   put_out_pp(yytext);
			   token=yylex();
			}
		     }
		  }
		  else {
		  }
		  do
		     token=yylex();
		  while (token==TWHITESPACE);
               }
	       else
		  error("No identifier following `ifdef.\n");
            }
	    break;
	 default:
            put_out_pp(yytext);
      }
   }
   while (token>0);
#endif
   if (verbose_mode || no_errors>0)
      info("Preprocessed %d lines with %d errors\n",line_no-1,no_errors);
   /*  prints out the identifier table for debugging
   for (cur_ID=0;cur_ID<last_macro_pp;cur_ID++)
      if (macro_table_pp[cur_ID].macro!=NULL)
	 printf("%d: %s ==> '%s'\n",cur_ID,macro_table_pp[cur_ID].ID,macro_table_pp[cur_ID].macro);
	 */
/*
{char line[1000];
rewind(ofile);
while (!feof(ofile)) {
   if (fscanf(ofile,"%s",line)==EOF) break;
   fprintf(ofile,"%s",line);
   }
}
*/
}
